#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import typing
from typing import Any, cast, Dict, List, Optional, TYPE_CHECKING, Union

import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.batch_job import (
    AbstractBatchJob,
    CsvSettings,
    IntakeSettings,
    OutputSettings,
    recognize_sourcedata,
    Schedule,
)
from datarobot.utils import get_id_from_response, pagination, to_api

from ..enums import DEFAULT_TIMEOUT, JOB_TYPE, QUEUE_STATUS
from ..utils import logger
from .api_object import APIObject

LOG = logger.get_logger(__name__)

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    from datarobot.models.deployment import Deployment

    DeploymentType = Union[str, Deployment]

    class CreatedBy(TypedDict):
        user_id: Optional[str]
        username: Optional[str]
        full_name: Optional[str]

    class MonitoringAggregation(TypedDict):
        retention_policy: str
        retention_value: int

    class PredictionColumnMap(TypedDict):
        class_name: str
        column_name: str

    class MonitoringColumns(TypedDict):
        predictions_columns: Optional[Union[str, PredictionColumnMap]]
        association_id_column: Optional[str]
        actuals_value_column: Optional[str]
        acted_upon_column: Optional[str]
        actuals_timestamp_column: Optional[str]

    class MonitoringOutputSettings(TypedDict):
        unique_row_identifier_columns: List[str]
        monitored_status_column: str

    class BatchMonitoringJobDict(TypedDict, total=False):
        """Batch monitoring job typed dict"""

        deployment_id: str
        intake_settings: IntakeSettings
        output_settings: Optional[OutputSettings]
        csv_settings: Optional[CsvSettings]
        num_concurrent: Optional[int]
        chunk_size: Optional[Union[int, str]]
        abort_on_error: bool
        download_timeout: Optional[int]
        download_read_timeout: Optional[int]
        upload_read_timeout: Optional[int]
        monitoring_aggregation: Optional[MonitoringAggregation]
        monitoring_columns: Optional[MonitoringColumns]
        monitoring_output_settings: Optional[MonitoringOutputSettings]


class BatchMonitoringJob(AbstractBatchJob):
    """
    A Batch Monitoring Job is used to monitor data sets outside DataRobot app.

    Attributes
    ----------
    id : str
        the id of the job
    """

    _job_spec = t.Dict(
        {
            t.Key("num_concurrent"): Int(),
            t.Key("deployment_id"): String(),
            t.Key("intake_settings", optional=True): t.Dict().allow_extra("*"),
            t.Key("output_settings", optional=True): t.Dict().allow_extra("*"),
            t.Key("monitoring_aggregation", optional=True): t.Dict().allow_extra("*"),
            t.Key("monitoring_columns", optional=True): t.Dict().allow_extra("*"),
            t.Key("monitoring_output_settings", optional=True): t.Dict().allow_extra("*"),
        }
    ).allow_extra("*")
    _links = t.Dict(
        {t.Key("download", optional=True): String(allow_blank=True), t.Key("self"): String()}
    ).allow_extra("*")
    _converter_extra = t.Dict(
        {
            t.Key("percentage_completed"): t.Float(),
            t.Key("elapsed_time_sec"): Int(),
            t.Key("links"): _links,
            t.Key("job_spec"): _job_spec,
            t.Key("status_details"): String(),
        }
    ).allow_extra("*")
    _converter_common = t.Dict(
        {
            t.Key("id", optional=True): String,
            t.Key("status", optional=True): t.Enum(
                QUEUE_STATUS.ABORTED,
                QUEUE_STATUS.COMPLETED,
                QUEUE_STATUS.RUNNING,
                QUEUE_STATUS.INITIALIZING,
                "FAILED",
            ),
            t.Key("project_id", optional=True): String,
            t.Key("is_blocked", optional=True): t.Bool,
        }
    )
    _monitoring_columns = t.Dict(
        {
            t.Key("predictions_columns", optional=True): t.List(
                t.Dict({t.Key("class_name"): t.String(), t.Key("column_name"): t.String()})
            )
            | t.String(),
            t.Key("association_id_column", optional=True): t.String(),
            t.Key("actuals_value_column", optional=True): t.String(),
            t.Key("acted_upon_column", optional=True): t.String(),
            t.Key("actuals_timestamp_column", optional=True): t.String(),
        }
    )
    _monitoring_output_settings = t.Dict(
        {
            t.Key("unique_row_identifier_columns", optional=True): t.List(t.String),
            t.Key("monitored_status_column", optional=True): t.String(),
        }
    )
    _monitoring_aggregation = t.Dict(
        {
            t.Key("retention_policy", optional=True): t.Enum("samples", "percentage"),
            t.Key("retention_value", optional=True, default=0): t.Int(),
        }
    )

    @classmethod
    def _job_type(cls) -> str:
        return cast(str, JOB_TYPE.BATCH_MONITORING)

    @classmethod
    def _job_path(cls, project_id: str, job_id: str) -> str:
        return f"batchJobs/{job_id}/"

    @classmethod
    def _jobs_path(cls) -> str:
        return "batchJobs/"

    @classmethod
    def _single_job_path(cls) -> str:
        return "batchMonitoring/"

    @classmethod
    def get(cls, project_id: Optional[str], job_id: str) -> BatchMonitoringJob:
        """Get batch monitoring job

        Attributes
        ----------
        job_id: str
            ID of batch job

        Returns
        -------
        BatchMonitoringJob
            Instance of BatchMonitoringJob
        """
        batch_job = super().get(project_id="", job_id=job_id)
        batch_job.id = job_id

        return batch_job

    def download(
        self, fileobj: typing.IO[Any], timeout: int = 120, read_timeout: int = 660
    ) -> None:
        """Downloads the results of a monitoring job as a CSV.

        Attributes
        ----------
        fileobj: A file-like object where the CSV monitoring results will be
            written to. Examples include an in-memory buffer
            (e.g., io.BytesIO) or a file on disk (opened for binary writing).

        timeout : int (optional, default 120)
            Seconds to wait for the download to become available.

            The download will not be available before the job has started processing.
            In case other jobs are occupying the queue, processing may not start
            immediately.

            If the timeout is reached, the job will be aborted and `RuntimeError`
            is raised.

            Set to -1 to wait infinitely.

        read_timeout : int (optional, default 660)
            Seconds to wait for the server to respond between chunks.
        """
        self._download(fileobj, timeout, read_timeout)

    @classmethod
    def run(
        cls,
        deployment: DeploymentType,
        intake_settings: Optional[IntakeSettings] = None,
        output_settings: Optional[OutputSettings] = None,
        csv_settings: Optional[CsvSettings] = None,
        num_concurrent: Optional[int] = None,
        chunk_size: Optional[Union[int, str]] = None,
        abort_on_error: bool = True,
        monitoring_aggregation: Optional[MonitoringAggregation] = None,
        monitoring_columns: Optional[MonitoringColumns] = None,
        monitoring_output_settings: Optional[MonitoringOutputSettings] = None,
        download_timeout: int = 120,
        download_read_timeout: int = 660,
        upload_read_timeout: int = DEFAULT_TIMEOUT.UPLOAD,
    ) -> BatchMonitoringJob:
        """
        Create new batch monitoring job, upload the dataset, and
        return a batch monitoring job.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for monitoring.

        intake_settings : dict
            A dict configuring how data is coming from. Supported options:

                - type : string, either `localFile`, `s3`, `azure`, `gcp`, `dataset`, `jdbc`
                  `snowflake`, `synapse`, `bigquery`, or `datasphere`

            Note that to pass a dataset, you not only need to specify the `type` parameter
            as `dataset`, but you must also set the `dataset` parameter as a
            `dr.Dataset` object.

            To monitor from a local file, add this parameter to the
            settings:

                - file : A file-like object, string path to a file or a
                  pandas.DataFrame of scoring data.

            To monitor from S3, add the next parameters to the settings:

                - url : string, the URL to score (e.g.: `s3://bucket/key`).
                - credential_id : string (optional).
                - endpoint_url : string (optional), any non-default endpoint
                  URL for S3 access (omit to use the default).

            (batch_monitoring_jdbc_creds_usage)=

            To monitor from JDBC, add the next parameters to the settings:

                - data_store_id : string, the ID of the external data store connected
                  to the JDBC data source (see
                  :ref:`Database Connectivity <database-connectivity-overview>`).
                - query : string (optional if `table`, `schema` and/or `catalog` is specified),
                  a self-supplied SELECT statement of the data set you wish to predict.
                - table : string (optional if `query` is specified),
                  the name of specified database table.
                - schema : string (optional if `query` is specified),
                  the name of specified database schema.
                - catalog : string  (optional if `query` is specified),
                  (new in v2.22) the name of specified database catalog.
                - fetch_size : int (optional),
                  Changing the `fetchSize` can be used to balance throughput and memory
                  usage.
                - credential_id : string (optional) the ID of the credentials holding
                  information about a user with read-access to the JDBC data source (see
                  :ref:`Credentials <credentials-api-doc>`).

            To monitor from Datasphere, add the following parameters to the settings:

                - `data_store_id` : string, the ID of the external data store connected to
                  the Datasphere data source (see
                  :ref:`Database Connectivity <database-connectivity-overview>`).
                - `table` : string,  the name of specified database table.
                - `schema` : string, the name of specified database schema.
                - `credential_id` : string, the ID of the credentials holding information about
                  a user with read-access to the Datasphere data source (see
                  :ref:`Credentials <credentials-api-doc>`).

        output_settings : dict (optional)
            A dict configuring how monitored data is to be saved. Supported
            options:

                - type : string, either `localFile`, `s3`, `azure`, `gcp`, `jdbc`,
                  `snowflake`, `synapse`, `bigquery`, or `datasphere`

            To save monitored data to a local file, add parameters to the
            settings:

                - path : string (optional), path to save the scored data
                  as CSV. If a path is not specified, you must download
                  the scored data yourself with `job.download()`.
                  If a path is specified, the call will block until the
                  job is done. if there are no other jobs currently
                  processing for the targeted prediction instance,
                  uploading, scoring, downloading will happen in parallel
                  without waiting for a full job to complete. Otherwise,
                  it will still block, but start downloading the scored
                  data as soon as it starts generating data. This is the
                  fastest method to get predictions.

            To save monitored data to S3, add the next parameters to the settings:

                - url : string, the URL for storing the results
                  (e.g.: `s3://bucket/key`).
                - credential_id : string (optional).
                - endpoint_url : string (optional), any non-default endpoint
                  URL for S3 access (omit to use the default).

            To save monitored data to JDBC, add the next parameters to the settings:

                - `data_store_id` : string, the ID of the external data store connected to
                  the JDBC data source (see
                  :ref:`Database Connectivity <database-connectivity-overview>`).
                - `table` : string,  the name of specified database table.
                - `schema` : string (optional), the name of specified database schema.
                - `catalog` : string (optional), (new in v2.22) the name of specified database
                  catalog.
                - `statement_type` : string, the type of insertion statement to create,
                  one of ``datarobot.enums.AVAILABLE_STATEMENT_TYPES``.
                - `update_columns` : list(string) (optional),  a list of strings containing
                  those column names to be updated in case `statement_type` is set to a
                  value related to update or upsert.
                - `where_columns` : list(string) (optional), a list of strings containing
                  those column names to be selected in case `statement_type` is set to a
                  value related to insert or update.
                - `credential_id` : string, the ID of the credentials holding information about
                  a user with write-access to the JDBC data source (see
                  :ref:`Credentials <credentials-api-doc>`).
                - `create_table_if_not_exists` : bool (optional), If no existing table is detected,
                  attempt to create it before writing data with the strategy defined in the
                  statementType parameter.

            To save monitored data to Datasphere, add the next parameters to the settings:

                - `data_store_id` : string, the ID of the external data store connected to
                  the Datasphere data source (see
                  :ref:`Database Connectivity <database-connectivity-overview>`).
                - `table` : string,  the name of specified database table.
                - `schema` : string, the name of specified database schema.
                - `credential_id` : string, the ID of the credentials holding information about
                  a user with write-access to the Datasphere data source (see
                  :ref:`Credentials <credentials-api-doc>`).

        csv_settings : dict (optional)
            CSV intake and output settings. Supported options:

            - `delimiter` : string (optional, default `,`), fields are delimited by
              this character. Use the string `tab` to denote TSV (TAB separated values).
              Must be either a one-character string or the string `tab`.
            - `quotechar` : string (optional, default `"`), fields containing the
              delimiter must be quoted using this character.
            - `encoding` : string (optional, default `utf-8`), encoding for the CSV
              files. For example (but not limited to): `shift_jis`, `latin_1` or
              `mskanji`.

        num_concurrent : int (optional)
            Number of concurrent chunks to score simultaneously. Defaults to
            the available number of cores of the deployment. Lower it to leave
            resources for real-time scoring.

        chunk_size : string or int (optional)
            Which strategy should be used to determine the chunk size.
            Can be either a named strategy or a fixed size in bytes.
            - auto: use fixed or dynamic based on flipper.
            - fixed: use 1MB for explanations, 5MB for regular requests.
            - dynamic: use dynamic chunk sizes.
            - int: use this many bytes per chunk.

        abort_on_error : boolean (optional)
             Default behavior is to abort the job if too many rows fail scoring. This will free
             up resources for other jobs that may score successfully. Set to `false` to
             unconditionally score every row no matter how many errors are encountered.
             Defaults to `True`.

        download_timeout : int (optional)
            .. versionadded:: 2.22

            If using localFile output, wait this many seconds for the download to become
            available. See `download()`.

        download_read_timeout : int (optional, default 660)
            .. versionadded:: 2.22

            If using localFile output, wait this many seconds for the server to respond
            between chunks.

        upload_read_timeout: int (optional, default 600)
            .. versionadded:: 2.28

            If using localFile intake, wait this many seconds for the server to respond
            after whole dataset upload.

        Returns
        -------
        BatchMonitoringJob
          Instance of BatchMonitoringJob

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> job_spec = {
            ...     "intake_settings": {
            ...         "type": "jdbc",
            ...         "data_store_id": "645043933d4fbc3215f17e34",
            ...         "catalog": "SANDBOX",
            ...         "table": "10kDiabetes_output_actuals",
            ...         "schema": "SCORING_CODE_UDF_SCHEMA",
            ...         "credential_id": "645043b61a158045f66fb329"
            ...     },
            >>>     "monitoring_columns": {
            ...         "predictions_columns": [
            ...             {
            ...                 "class_name": "True",
            ...                 "column_name": "readmitted_True_PREDICTION"
            ...             },
            ...             {
            ...                 "class_name": "False",
            ...                 "column_name": "readmitted_False_PREDICTION"
            ...             }
            ...         ],
            ...         "association_id_column": "rowID",
            ...         "actuals_value_column": "ACTUALS"
            ...     }
            ... }
            >>> deployment_id = "foobar"
            >>> job = dr.BatchMonitoringJob.run(deployment_id, **job_spec)
            >>> job.wait_for_completion()
        """
        try:
            deployment_id = cast("Deployment", deployment).id
        except AttributeError:
            deployment_id = cast(str, deployment)

        job_data: BatchMonitoringJobDict = {"deployment_id": deployment_id}

        if num_concurrent is not None:
            job_data["num_concurrent"] = int(num_concurrent)

        if chunk_size is not None:
            job_data["chunk_size"] = chunk_size

        if not abort_on_error:
            job_data["abort_on_error"] = bool(abort_on_error)

        if csv_settings is not None:
            cls._csv_settings.check(csv_settings)
            job_data["csv_settings"] = csv_settings

        if monitoring_columns is not None:
            cls._monitoring_columns.check(monitoring_columns)
            job_data["monitoring_columns"] = monitoring_columns

        if monitoring_aggregation is not None:
            cls._monitoring_aggregation.check(monitoring_aggregation)
            job_data["monitoring_aggregation"] = monitoring_aggregation

        if monitoring_output_settings is not None:
            cls._monitoring_output_settings.check(monitoring_output_settings)
            job_data["monitoring_output_settings"] = monitoring_output_settings

        # validate input settings, return a copy not original
        intake_settings = cls.validate_intake_settings(intake_settings)
        intake_file = None
        if intake_settings["type"] == "localFile":
            intake_file = recognize_sourcedata(intake_settings.pop("file"), "prediction.csv")
        job_data["intake_settings"] = intake_settings

        # validate output settings, return a copy not original
        output_file = None
        if output_settings:
            output_settings = cls.validate_output_settings(output_settings)
            if output_settings["type"] == "localFile":
                if output_settings.get("path") is not None:
                    output_file = open(  # pylint: disable=consider-using-with
                        output_settings.pop("path"), "wb"  # type: ignore[arg-type]
                    )
            job_data["output_settings"] = output_settings

        payload = cast(Dict[str, Any], to_api(job_data))
        response = cls._client.post(url=cls._single_job_path(), json=payload)

        job_response = response.json()
        job_id = get_id_from_response(response)

        upload_thread = None

        if intake_file is not None:
            upload_thread = cls._upload_intake_file(
                intake_file, intake_settings, job_response, upload_read_timeout, output_file
            )

        job = BatchMonitoringJob.get(None, job_id)

        if output_file is not None:
            # We must download the result to `output_file`
            # And clean up any thread we spawned during uploading
            try:
                job.download(
                    output_file, timeout=download_timeout, read_timeout=download_read_timeout
                )
            finally:
                output_file.close()
                if upload_thread is not None:
                    upload_thread.join()

        return job

    def cancel(self, ignore_404_errors: bool = False) -> None:
        """
        Cancel this job. If this job has not finished running, it will be
        removed and canceled.
        """
        self._delete(ignore_404_errors)

    def get_status(self) -> Any:
        """Get status of batch monitoring job

        Returns
        -------
        BatchMonitoringJob status data
            Dict with job status
        """
        return self._get_status()


class BatchMonitoringJobDefinition(APIObject):  # pylint: disable=missing-class-docstring
    _path = "batchMonitoringJobDefinitions/"

    _user = t.Dict(
        {
            t.Key("username"): String(),
            t.Key("full_name", optional=True): String(),
            t.Key("user_id"): String(),
        }
    ).allow_extra("*")

    _schedule = t.Dict(
        {
            t.Key("day_of_week"): t.List(t.Or(String, Int)),
            t.Key("month"): t.List(t.Or(String, Int)),
            t.Key("hour"): t.List(t.Or(String, Int)),
            t.Key("minute"): t.List(t.Or(String, Int)),
            t.Key("day_of_month"): t.List(t.Or(String, Int)),
        }
    ).allow_extra("*")

    _converter = t.Dict(
        {
            t.Key("id"): String,
            t.Key("name"): String,
            t.Key("enabled"): t.Bool(),
            t.Key("schedule", optional=True): _schedule,
            t.Key("batch_monitoring_job"): BatchMonitoringJob._job_spec,
            t.Key("created"): String(),
            t.Key("updated"): String(),
            t.Key("created_by"): _user,
            t.Key("updated_by"): _user,
            t.Key("last_failed_run_time", optional=True): String(),
            t.Key("last_successful_run_time", optional=True): String(),
            t.Key("last_successful_run_time", optional=True): String(),
            t.Key("last_scheduled_run_time", optional=True): String(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        enabled: Optional[bool] = None,
        schedule: Optional[Schedule] = None,
        batch_monitoring_job: Optional[BatchMonitoringJobDict] = None,
        created: Optional[str] = None,
        updated: Optional[str] = None,
        created_by: Optional[CreatedBy] = None,
        updated_by: Optional[CreatedBy] = None,
        last_failed_run_time: Optional[str] = None,
        last_successful_run_time: Optional[str] = None,
        last_started_job_status: Optional[str] = None,
        last_scheduled_run_time: Optional[str] = None,
    ) -> None:
        self.id = id
        self.name = name
        self.enabled = enabled
        self.schedule = schedule
        self.batch_monitoring_job = batch_monitoring_job

        self.created = created
        self.updated = updated
        self.created_by = created_by
        self.updated_by = updated_by

        self.last_failed_run_time = last_failed_run_time
        self.last_successful_run_time = last_successful_run_time
        self.last_started_job_status = last_started_job_status
        self.last_scheduled_run_time = last_scheduled_run_time

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self.id == other.id

    @classmethod
    def get(cls, batch_monitoring_job_definition_id: str) -> BatchMonitoringJobDefinition:
        """Get batch monitoring job definition

        Attributes
        ----------
        batch_monitoring_job_definition_id: str
            ID of batch monitoring job definition

        Returns
        -------
        BatchMonitoringJobDefinition
            Instance of BatchMonitoringJobDefinition

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> definition = dr.BatchMonitoringJobDefinition.get('5a8ac9ab07a57a0001be501f')
            >>> definition
            BatchMonitoringJobDefinition(60912e09fd1f04e832a575c1)
        """

        return cls.from_location(f"{cls._path}{batch_monitoring_job_definition_id}/")

    @classmethod
    def list(cls) -> List[BatchMonitoringJobDefinition]:
        """
        Get job all monitoring job definitions

        Returns
        -------
        List[BatchMonitoringJobDefinition]
            List of job definitions the user has access to see

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> definition = dr.BatchMonitoringJobDefinition.list()
            >>> definition
            [
                BatchMonitoringJobDefinition(60912e09fd1f04e832a575c1),
                BatchMonitoringJobDefinition(6086ba053f3ef731e81af3ca)
            ]
        """

        return list(
            cls.from_server_data(item) for item in pagination.unpaginate(cls._path, {}, cls._client)
        )

    @classmethod
    def create(
        cls,
        enabled: bool,
        batch_monitoring_job: BatchMonitoringJobDict,
        name: Optional[str] = None,
        schedule: Optional[Schedule] = None,
    ) -> BatchMonitoringJobDefinition:
        """
        Creates a new batch monitoring job definition to be run either at scheduled interval or as
        a manual run.

        Attributes
        ----------
        enabled : bool (default False)
            Whether the definition should be active on a scheduled basis. If True,
            `schedule` is required.

        batch_monitoring_job: dict
            The job specifications for your batch monitoring job.
            It requires the same job input parameters as used with BatchMonitoringJob

        name : string (optional)
            The name you want your job to be identified with. Must be unique across the
            organization's existing jobs.
            If you don't supply a name, a random one will be generated for you.

        schedule : dict (optional)
            The ``schedule`` payload defines at what intervals the job should run, which can be
            combined in various ways to construct complex scheduling terms if needed. In all
            the elements in the objects, you can supply either an asterisk ``["*"]`` denoting
            "every" time denomination or an array of integers (e.g. ``[1, 2, 3]``) to define
            a specific interval.

            The ``schedule`` payload is split up in the following items:

            **Minute:**

            The minute(s) of the day that the job will run. Allowed values are either ``["*"]``
            meaning every minute of the day or ``[0 ... 59]``

            **Hour:**
            The hour(s) of the day that the job will run. Allowed values are either ``["*"]``
            meaning every hour of the day or ``[0 ... 23]``.

            **Day of Month:**
            The date(s) of the month that the job will run. Allowed values are either
            ``[1 ... 31]`` or ``["*"]`` for all days of the month. This field is additive with
            ``dayOfWeek``, meaning the job will run both on the date(s) defined in this field
            and the day specified by ``dayOfWeek`` (for example, dates 1st, 2nd, 3rd, plus every
            Tuesday). If ``dayOfMonth`` is set to ``["*"]`` and ``dayOfWeek`` is defined,
            the scheduler will trigger on every day of the month that matches ``dayOfWeek``
            (for example, Tuesday the 2nd, 9th, 16th, 23rd, 30th).
            Invalid dates such as February 31st are ignored.

            **Month:**
            The month(s) of the year that the job will run. Allowed values are either
            ``[1 ... 12]`` or ``["*"]`` for all months of the year. Strings, either
            3-letter abbreviations or the full name of the month, can be used
            interchangeably (e.g., "jan" or "october").
            Months that are not compatible with ``dayOfMonth`` are ignored, for example
            ``{"dayOfMonth": [31], "month":["feb"]}``

            **Day of Week:**
            The day(s) of the week that the job will run. Allowed values are ``[0 .. 6]``,
            where (Sunday=0), or ``["*"]``, for all days of the week. Strings, either 3-letter
            abbreviations or the full name of the day, can be used interchangeably
            (e.g., "sunday", "Sunday", "sun", or "Sun", all map to ``[0]``.
            This field is additive with ``dayOfMonth``, meaning the job will run both on the
            date specified by ``dayOfMonth`` and the day defined in this field.

        Returns
        -------
        BatchMonitoringJobDefinition
            Instance of BatchMonitoringJobDefinition

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> job_spec = {
            ...    "num_concurrent": 4,
            ...    "deployment_id": "foobar",
            ...    "intake_settings": {
            ...        "url": "s3://foobar/123",
            ...        "type": "s3",
            ...        "format": "csv"
            ...    },
            ...    "output_settings": {
            ...        "url": "s3://foobar/123",
            ...        "type": "s3",
            ...        "format": "csv"
            ...    },
            ...}
            >>> schedule = {
            ...    "day_of_week": [
            ...        1
            ...    ],
            ...    "month": [
            ...        "*"
            ...    ],
            ...    "hour": [
            ...        16
            ...    ],
            ...    "minute": [
            ...        0
            ...    ],
            ...    "day_of_month": [
            ...        1
            ...    ]
            ...}
            >>> definition = BatchMonitoringJobDefinition.create(
            ...    enabled=False,
            ...    batch_monitoring_job=job_spec,
            ...    name="some_definition_name",
            ...    schedule=schedule
            ... )
            >>> definition
            BatchMonitoringJobDefinition(60912e09fd1f04e832a575c1)
        """

        BatchMonitoringJob._job_spec.check(batch_monitoring_job)

        job_spec = cast(Dict[str, Any], to_api(batch_monitoring_job))

        payload: Dict[str, Any] = {
            "name": name,
            "enabled": enabled,
        }

        if schedule:
            payload["schedule"] = to_api(schedule)

        payload.update(**job_spec)

        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    def update(
        self,
        enabled: bool,
        batch_monitoring_job: Optional[BatchMonitoringJobDict] = None,
        name: Optional[str] = None,
        schedule: Optional[Schedule] = None,
    ) -> BatchMonitoringJobDefinition:
        """
        Updates a job definition with the changed specs.

        Takes the same input as :func:`~BatchMonitoringJobDefinition.create`

        Attributes
        ----------
        enabled : bool (default False)
            Same as ``enabled`` in :func:`~BatchMonitoringJobDefinition.create`.

        batch_monitoring_job: dict
            Same as ``batch_monitoring_job`` in :func:`~BatchMonitoringJobDefinition.create`.

        name : string (optional)
            Same as ``name`` in :func:`~BatchMonitoringJobDefinition.create`.

        schedule : dict
            Same as ``schedule`` in :func:`~BatchMonitoringJobDefinition.create`.

        Returns
        -------
        BatchMonitoringJobDefinition
            Instance of the updated BatchMonitoringJobDefinition

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> job_spec = {
            ...    "num_concurrent": 5,
            ...    "deployment_id": "foobar_new",
            ...    "intake_settings": {
            ...        "url": "s3://foobar/123",
            ...        "type": "s3",
            ...        "format": "csv"
            ...    },
            ...    "output_settings": {
            ...        "url": "s3://foobar/123",
            ...        "type": "s3",
            ...        "format": "csv"
            ...    },
            ...}
            >>> schedule = {
            ...    "day_of_week": [
            ...        1
            ...    ],
            ...    "month": [
            ...        "*"
            ...    ],
            ...    "hour": [
            ...        "*"
            ...    ],
            ...    "minute": [
            ...        30, 59
            ...    ],
            ...    "day_of_month": [
            ...        1, 2, 6
            ...    ]
            ...}
            >>> definition = BatchMonitoringJobDefinition.create(
            ...    enabled=False,
            ...    batch_monitoring_job=job_spec,
            ...    name="updated_definition_name",
            ...    schedule=schedule
            ... )
            >>> definition
            BatchMonitoringJobDefinition(60912e09fd1f04e832a575c1)
        """
        payload: Dict[str, Any] = {
            "enabled": enabled,
        }

        if name:
            payload["name"] = name

        if schedule:
            payload["schedule"] = to_api(schedule)

        if batch_monitoring_job:
            BatchMonitoringJob._job_spec.check(batch_monitoring_job)
            job_spec = cast(Dict[str, Any], to_api(batch_monitoring_job))
            payload.update(**job_spec)

        return self.from_server_data(
            self._client.patch(f"{self._path}{self.id}", data=payload).json()
        )

    def run_on_schedule(self, schedule: Schedule) -> BatchMonitoringJobDefinition:
        """
        Sets the run schedule of an already created job definition.

        If the job was previously not enabled, this will also set the job to enabled.

        Attributes
        ----------
        schedule : dict
            Same as ``schedule`` in :func:`~BatchMonitoringJobDefinition.create`.

        Returns
        -------
        BatchMonitoringJobDefinition
            Instance of the updated BatchMonitoringJobDefinition with the new / updated schedule.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> definition = dr.BatchMonitoringJobDefinition.create('...')
            >>> schedule = {
            ...    "day_of_week": [
            ...        1
            ...    ],
            ...    "month": [
            ...        "*"
            ...    ],
            ...    "hour": [
            ...        "*"
            ...    ],
            ...    "minute": [
            ...        30, 59
            ...    ],
            ...    "day_of_month": [
            ...        1, 2, 6
            ...    ]
            ...}
            >>> definition.run_on_schedule(schedule)
            BatchMonitoringJobDefinition(60912e09fd1f04e832a575c1)
        """

        payload = {
            "enabled": True,
            "schedule": to_api(schedule),
        }

        return self.from_server_data(
            self._client.patch(f"{self._path}{self.id}", data=payload).json()
        )

    def run_once(self) -> BatchMonitoringJob:
        """
        Manually submits a batch monitoring job to the queue, based off of an already
        created job definition.

        Returns
        -------
        BatchMonitoringJob
          Instance of BatchMonitoringJob

        Examples
        --------
        .. code-block:: python

          >>> import datarobot as dr
          >>> definition = dr.BatchMonitoringJobDefinition.create('...')
          >>> job = definition.run_once()
          >>> job.wait_for_completion()
        """

        definition = self.from_location(f"{self._path}{self.id}/")

        payload = {"jobDefinitionId": definition.id}

        response = self._client.post(
            f"{BatchMonitoringJob._jobs_path()}fromJobDefinition/", data=payload
        ).json()

        job_id = response["id"]
        return BatchMonitoringJob.get(None, job_id)

    def delete(self) -> None:
        """
        Deletes the job definition and disables any future schedules of this job if any.
        If a scheduled job is currently running, this will not be cancelled.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> definition = dr.BatchMonitoringJobDefinition.get('5a8ac9ab07a57a0001be501f')
            >>> definition.delete()
        """

        self._client.delete(f"{self._path}{self.id}/")



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import os
from typing import List, Optional

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject

from ..enums import DEFAULT_MAX_WAIT
from ..utils import deprecation_warning, get_id_from_location
from ..utils.waiters import wait_for_async_resolution


class Connector(APIObject):
    """A connector

    Attributes
    ----------
    id : str
        the id of the connector.
    creator_id : str
        the id of the user who created the connector.
    base_name : str
        the file name of the jar file.
    canonical_name : str
        the user-friendly name of the connector.
    configuration_id : str
        the id of the configuration of the connector.
    """

    _path = "externalConnectors/"
    _file_upload_path = "externalDataDriverFile/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("creator_id"): String(),
            t.Key("configuration_id"): String(),
            t.Key("base_name", optional=True): t.Or(String, t.Null),
            t.Key("canonical_name"): String(),
            t.Key("connector_type", optional=True): t.Or(String, t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: Optional[str] = None,
        creator_id: Optional[str] = None,
        configuration_id: Optional[str] = None,
        base_name: Optional[str] = None,
        canonical_name: Optional[str] = None,
        connector_type: Optional[str] = None,
    ):
        self._id = id
        self._creator_id = creator_id
        self._configuration_id = configuration_id
        self._base_name = base_name
        self._canonical_name = canonical_name
        self._connector_type = connector_type

    @classmethod
    def list(cls) -> List[Connector]:
        """
        Returns list of available connectors.

        Returns
        -------
        connectors : list of Connector instances
            contains a list of available connectors.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> connectors = dr.Connector.list()
            >>> connectors
            [Connector('ADLS Gen2 Connector'), Connector('S3 Connector')]
        """
        r_data = cls._client.get(cls._path).json()
        return [cls.from_server_data(item) for item in r_data["data"]]

    @classmethod
    def get(cls, connector_id: str) -> Connector:
        """
        Gets the connector.

        Parameters
        ----------
        connector_id : str
            the identifier of the connector.

        Returns
        -------
        connector : Connector
            the required connector.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> connector = dr.Connector.get('5fe1063e1c075e0245071446')
            >>> connector
            Connector('ADLS Gen2 Connector')
        """
        return cls.from_location(f"{cls._path}{connector_id}/")

    @classmethod
    def create(
        cls, file_path: Optional[str] = None, connector_type: Optional[str] = None
    ) -> Connector:
        """
        Creates the connector from a jar file. Only available to admin users.

        Parameters
        ----------
        file_path : str
            (Deprecated in version v3.6)
            the file path on file system file_path(s) for the java-based connector.
        connector_type: str
            The type of the native connector to create

        Returns
        -------
        connector : Connector
            the created connector.

        Raises
        ------
        ClientError
            raised if user is not granted for `Can manage connectors` feature

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> connector = dr.Connector.create('/tmp/connector-adls-gen2.jar')
            >>> connector
            Connector('ADLS Gen2 Connector')
        """
        if file_path:
            deprecation_warning(
                subject="Parameter `file_path` in method Connector.create",
                deprecated_since_version="v3.6",
                will_remove_version="v3.7",
                message="Java-based connectors are deprecated.",
            )

            base_name = file_path.split(os.sep)[-1]
            resp = cls._client.build_request_with_file(
                "POST", cls._file_upload_path, base_name, file_path=file_path
            ).json()
            local_url = resp["localUrl"]

            payload = {"local_url": local_url, "base_name": base_name}
            resp = cls._client.post(cls._path, data=payload)
        elif connector_type:
            payload = {"connector_type": connector_type}
            resp = cls._client.post(cls._path, data=payload)
        else:
            raise ValueError("Must pass either file_path or connector_type")

        finished_location = wait_for_async_resolution(
            cls._client, resp.headers["Location"], max_wait=DEFAULT_MAX_WAIT
        )
        connector_id = get_id_from_location(finished_location)
        return cls.get(connector_id)

    def update(self, file_path: str) -> Connector:
        """
        Updates the connector with new jar file. Only available to admin users.

        Parameters
        ----------
        file_path : str
            (Deprecated in version v3.6)
            the file path on file system file_path(s) for the java-based connector.

        Returns
        -------
        connector : Connector
            the updated connector.

        Raises
        ------
        ClientError
            raised if user is not granted for `Can manage connectors` feature

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> connector = dr.Connector.get('5fe1063e1c075e0245071446')
            >>> connector.base_name
            'connector-adls-gen2.jar'
            >>> connector.update('/tmp/connector-s3.jar')
            >>> connector.base_name
            'connector-s3.jar'
        """
        deprecation_warning(
            subject="Parameter `file_path` in method Connector.update",
            deprecated_since_version="v3.6",
            will_remove_version="v3.7",
            message="Java-based connectors are deprecated.",
        )
        base_name = file_path.split(os.sep)[-1]
        resp = self._client.build_request_with_file(
            "POST", self._file_upload_path, base_name, file_path=file_path
        ).json()
        local_url = resp["localUrl"]

        payload = {"local_url": local_url, "base_name": base_name}
        resp = self._client.patch(f"{self._path}{self.id}/", data=payload)
        wait_for_async_resolution(self._client, resp.headers["Location"], max_wait=DEFAULT_MAX_WAIT)

        data = self._client.get(f"{self._path}{self.id}/").json()

        self._configuration_id = data["configurationId"]
        self._base_name = data["baseName"]
        self._canonical_name = data["canonicalName"]
        return self

    def delete(self) -> None:
        """
        Removes the connector. Only available to admin users.

        Raises
        ------
        ClientError
            raised if user is not granted for `Can manage connectors` feature
        """
        self._client.delete(f"{self._path}{self.id}/")

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def creator(self) -> Optional[str]:
        return self._creator_id

    @property
    def configuration_id(self) -> Optional[str]:
        return self._configuration_id

    @property
    def base_name(self) -> Optional[str]:
        return self._base_name

    @property
    def canonical_name(self) -> Optional[str]:
        return self._canonical_name

    @property
    def connector_type(self) -> Optional[str]:
        return self._connector_type

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.canonical_name or self.id}')"



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Optional, TYPE_CHECKING, Union

import trafaret as t

from datarobot import errors
from datarobot._compat import String
from datarobot.helpers.feature_discovery import SecondaryDataset
from datarobot.models.api_object import APIObject
from datarobot.utils import parse_time

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    from datarobot.helpers.feature_discovery import SecondaryDatasetDict
    from datarobot.models.api_object import ServerDataType

    class DatasetConfigurationDict(TypedDict):
        feature_engineering_graph_id: Optional[str]
        secondary_datasets: Optional[List[SecondaryDatasetDict]]

    class StoredCredentials(TypedDict):
        credential_id: str
        catalog_version_id: str
        url: Optional[str]

    class SecondaryDatasetConfigurationsDict(TypedDict):
        """Secondary Dataset Config Typed Dict"""

        id: str
        project_id: str
        config: Optional[List[DatasetConfigurationDict]]
        secondary_datasets: Optional[List[SecondaryDatasetDict]]
        name: Optional[str]
        creator_full_name: Optional[str]
        creator_user_id: Optional[str]
        created: Optional[datetime]
        featurelist_id: Optional[str]
        credential_ids: Optional[StoredCredentials]
        is_default: Optional[bool]
        project_version: Optional[str]


class DatasetConfiguration:
    """Specify a dataset configuration

    .. versionadded:: v2.20

    Attributes
    ----------
    feature_engineering_graph_id: str
        Id of the feature engineering graph
    secondary_datasets: list of SecondaryDataset
        List of secondary datasets
    """

    def __init__(
        self,
        feature_engineering_graph_id: Optional[str] = None,
        secondary_datasets: Optional[List[SecondaryDataset]] = None,
    ) -> None:
        self.feature_engineering_graph_id = feature_engineering_graph_id
        self.secondary_datasets = secondary_datasets

    def to_dict(self) -> DatasetConfigurationDict:
        secondary_datasets = self.secondary_datasets if self.secondary_datasets is not None else []
        return {
            "feature_engineering_graph_id": self.feature_engineering_graph_id,
            "secondary_datasets": [dataset.to_dict() for dataset in secondary_datasets],
        }


class DatasetsCredentials:
    """Credentials of the JDBC/ dynamic based secondary datasets

    .. versionadded:: v2.23

    Attributes
    ----------
    credential_id: str
        Id of credential store
    catalog_version_id: str
        Catalog version Id of the secondary dataset
    url: Optional[str]
        JDBC connection URL for the dataset
    """

    def __init__(
        self,
        credential_id: Optional[str] = None,
        catalog_version_id: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        self.credential_id = credential_id
        self.catalog_version_id = catalog_version_id
        self.url = url

    def to_dict(self) -> Dict[str, Optional[str]]:
        return {
            "credential_id": self.credential_id,
            "catalog_version_id": self.catalog_version_id,
            "url": self.url,
        }


class SecondaryDatasetConfigurations(APIObject):
    """Create secondary dataset configurations for a given project

    .. versionadded:: v2.20

    Attributes
    ----------
    id : str
        Id of this secondary dataset configuration
    project_id : str
        Id of the associated project.
    config: list of DatasetConfiguration (Deprecated in version v2.23)
        List of secondary dataset configurations
    secondary_datasets: list of SecondaryDataset (new in v2.23)
        List of secondary datasets (secondaryDataset)
    name: str
        Verbose name of the SecondaryDatasetConfig. null if it wasn't specified.
    created: datetime.datetime
        DR-formatted datetime. null for legacy (before DR 6.0) db records.
    creator_user_id: str
        Id of the user created this config.
    creator_full_name: str
        fullname or email of the user created this config.
    featurelist_id: Optional[str]
        Id of the feature list. null if it wasn't specified.
    credential_ids: Optional[list of DatasetsCredentials]
        credentials used by the secondary datasets if the datasets used
        in the configuration are from datasource
    is_default: Optional[bool]
        Boolean flag if default config created during feature discovery aim
    project_version: Optional[str]
        Version of project when its created (Release version)
    """

    _base_url = "projects/{}/secondaryDatasetsConfigurations/"
    _stored_credentials = t.Dict(
        {
            t.Key("credential_id"): String,
            t.Key("catalog_version_id"): String,
            t.Key("url", optional=True): t.Or(String, t.Null),
        }
    )
    _secondary_dataset_converter = t.Dict(
        {
            t.Key("identifier"): String(min_length=3, max_length=20),
            t.Key("catalog_version_id"): String,
            t.Key("catalog_id"): String,
            t.Key("snapshot_policy", optional=True, default="latest"): t.Enum(
                "latest", "specified", "dynamic"
            ),
        }
    )

    _dataset_configuration_converter = t.Dict(
        {
            t.Key("feature_engineering_graph_id"): String(),
            t.Key("secondary_datasets"): t.List(_secondary_dataset_converter),
        }
    ).ignore_extra("*")

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("project_id"): String(),
            t.Key("config", optional=True): t.List(_dataset_configuration_converter),
            t.Key("secondary_datasets"): t.List(_secondary_dataset_converter),
            t.Key("name", optional=True): t.Or(String, t.Null),
            t.Key("creator_full_name", optional=True): String(),
            t.Key("creator_user_id", optional=True): String(),
            t.Key("created", optional=True): parse_time,
            t.Key("featurelist_id", optional=True): t.Or(String, t.Null),
            t.Key("credential_ids", optional=True): t.List(_stored_credentials),
            t.Key("is_default", optional=True): t.Bool,
            t.Key("project_version", optional=True): t.Or(String, t.Null),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        project_id: str,
        config: Optional[List[DatasetConfiguration]] = None,
        secondary_datasets: Optional[List[SecondaryDataset]] = None,
        name: Optional[str] = None,
        creator_full_name: Optional[str] = None,
        creator_user_id: Optional[str] = None,
        created: Optional[datetime] = None,
        featurelist_id: Optional[str] = None,
        credential_ids: Optional[StoredCredentials] = None,
        is_default: Optional[bool] = None,
        project_version: Optional[str] = None,
    ) -> None:
        self.id = id
        self.project_id = project_id
        self.config = config
        self.name = name
        self.secondary_datasets = secondary_datasets
        self.creator_full_name = creator_full_name
        self.creator_user_id = creator_user_id
        self.created = created
        self.featurelist_id = featurelist_id
        self.credential_ids = credential_ids
        self.is_default = is_default
        self.project_version = project_version

    def to_dict(self) -> SecondaryDatasetConfigurationsDict:
        config, secondary_datasets = None, None
        if self.config:
            config = [c.to_dict() for c in self.config]
        if self.secondary_datasets:
            secondary_datasets = [sc.to_dict() for sc in self.secondary_datasets]
        return {
            "id": self.id,
            "project_id": self.project_id,
            "config": config,
            "name": self.name,
            "secondary_datasets": secondary_datasets,
            "creator_full_name": self.creator_full_name,
            "creator_user_id": self.creator_user_id,
            "created": self.created,
            "featurelist_id": self.featurelist_id,
            "credential_ids": self.credential_ids,
            "is_default": self.is_default,
            "project_version": self.project_version,
        }

    @classmethod
    def from_data(cls, data: ServerDataType) -> SecondaryDatasetConfigurations:
        checked = cls._converter.check(data)
        safe_data = cls._filter_data(checked)

        id = safe_data.get("id", None)
        project_id = safe_data.get("project_id", None)
        name = safe_data.get("name", None)
        secondary_datasets = safe_data.get("secondary_datasets", None)
        creator_full_name = safe_data.get("creator_full_name", None)
        creator_user_id = safe_data.get("creator_user_id", None)
        created = safe_data.get("created", None)
        featurelist_id = safe_data.get("featurelist_id", None)
        credential_ids = safe_data.get("credential_ids", None)
        is_default = safe_data.get("is_default", None)
        project_version = safe_data.get("project_version", None)
        dataset_configs = safe_data.get("config", [])

        conf_list = []
        for conf in dataset_configs:
            graph_id = conf.get("feature_engineering_graph_id", None)
            datasets = conf.get("secondary_datasets", None)
            dataset_list = []
            for d in datasets:
                dataset = SecondaryDataset(**d)
                dataset_list.append(dataset)
            dataset_configuration = DatasetConfiguration(graph_id, dataset_list)
            conf_list.append(dataset_configuration)

        secondary_datasets_list = []
        for d in secondary_datasets:
            dataset = SecondaryDataset(**d)
            secondary_datasets_list.append(dataset)
        return SecondaryDatasetConfigurations(
            id=id,
            project_id=project_id,
            config=conf_list,
            secondary_datasets=secondary_datasets_list,
            name=name,
            creator_full_name=creator_full_name,
            creator_user_id=creator_user_id,
            created=created,
            featurelist_id=featurelist_id,
            credential_ids=credential_ids,
            is_default=is_default,
            project_version=project_version,
        )

    @classmethod
    def create(
        cls,
        project_id: str,
        secondary_datasets: List[SecondaryDataset],
        name: str,
        featurelist_id: Optional[str] = None,
    ) -> SecondaryDatasetConfigurations:
        """create secondary dataset configurations

        .. versionadded:: v2.20

        Parameters
        ----------
        project_id : str
            id of the associated project.
        secondary_datasets: list of SecondaryDataset (New in version v2.23)
            list of secondary datasets used by the configuration
            each element is a ``datarobot.helpers.feature_discovery.SecondaryDataset``
        name: str (New in version v2.23)
            Name of the secondary datasets configuration
        featurelist_id: str, or None (New in version v2.23)
            Id of the featurelist

        Returns
        -------
        an instance of SecondaryDatasetConfigurations

        Raises
        ------
        ClientError
            raised if incorrect configuration parameters are provided

        Examples
        --------
        .. code-block:: python

            profile_secondary_dataset = dr.SecondaryDataset(
                identifier='profile',
                catalog_id='5ec4aec1f072bc028e3471ae',
                catalog_version_id='5ec4aec2f072bc028e3471b1',
                snapshot_policy='latest'
            )

            transaction_secondary_dataset = dr.SecondaryDataset(
                identifier='transaction',
                catalog_id='5ec4aec268f0f30289a03901',
                catalog_version_id='5ec4aec268f0f30289a03900',
                snapshot_policy='latest'
            )

            secondary_datasets = [profile_secondary_dataset, transaction_secondary_dataset]
            new_secondary_dataset_config = dr.SecondaryDatasetConfigurations.create(
                project_id=project.id,
                name='My config',
                secondary_datasets=secondary_datasets
            )


            >>> new_secondary_dataset_config.id
            '5fd1e86c589238a4e635e93d'
        """
        if not project_id:
            raise errors.ClientError("project_id cannot be None or empty", 422)
        if not secondary_datasets:
            raise errors.ClientError(
                "secondary_datasets cannot be None or empty",
                422,
            )

        url = cls._base_url.format(project_id)
        payload_secondary_datasets = []
        for sec_dataset in secondary_datasets:
            if isinstance(sec_dataset, SecondaryDataset):
                payload_secondary_datasets.append(sec_dataset.to_payload())
            else:
                payload_secondary_datasets.append(sec_dataset)  # type: ignore[unreachable]

        payload = {
            "name": name,
            "featurelistId": featurelist_id,
            "secondaryDatasets": payload_secondary_datasets,
        }
        response = cls._client.post(url, data=payload)

        status = response.status_code
        if status == 201:
            return cls.from_server_data(response.json())
        else:
            error_msg = response.json().get(
                "message", "error in processing secondary dataset configuration"
            )
            raise errors.ClientError(
                error_msg + f" with server returned status {status}",
                status,
            )

    def delete(self) -> None:
        """Removes the Secondary datasets configuration

        .. versionadded:: v2.21

        Raises
        ------
        ClientError
            Raised if an invalid or already deleted secondary dataset config id is provided

        Examples
        --------
        .. code-block:: python

            # Deleting with a valid secondary_dataset_config id
            status_code = dr.SecondaryDatasetConfigurations.delete(some_config_id)
            status_code
            >>> 204
        """
        url = self._base_url.format(self.project_id)
        self._client.delete(f"{url}{self.id}/")

    def get(self) -> SecondaryDatasetConfigurations:
        """Retrieve a single secondary dataset configuration for a given id

        .. versionadded:: v2.21

        Returns
        -------
        secondary_dataset_configurations : SecondaryDatasetConfigurations
            The requested secondary dataset configurations

        Examples
        --------
        .. code-block:: python

            config_id = '5fd1e86c589238a4e635e93d'
            secondary_dataset_config = dr.SecondaryDatasetConfigurations(id=config_id).get()
            >>> secondary_dataset_config
            {
                 'created': datetime.datetime(2020, 12, 9, 6, 16, 22, tzinfo=tzutc()),
                 'creator_full_name': u'abc@datarobot.com',
                 'creator_user_id': u'asdf4af1gf4bdsd2fba1de0a',
                 'credential_ids': None,
                 'featurelist_id': None,
                 'id': u'5fd1e86c589238a4e635e93d',
                 'is_default': True,
                 'name': u'My config',
                 'project_id': u'5fd06afce2456ec1e9d20457',
                 'project_version': None,
                 'secondary_datasets': [
                        {
                            'snapshot_policy': u'latest',
                            'identifier': u'profile',
                            'catalog_version_id': u'5fd06b4af24c641b68e4d88f',
                            'catalog_id': u'5fd06b4af24c641b68e4d88e'
                        },
                        {
                            'snapshot_policy': u'dynamic',
                            'identifier': u'transaction',
                            'catalog_version_id': u'5fd1e86c589238a4e635e98e',
                            'catalog_id': u'5fd1e86c589238a4e635e98d'
                        }
                 ]
            }
        """
        assert self.project_id
        url = self._base_url.format(self.project_id)
        return self.from_location(f"{url}{self.id}/")

    @classmethod
    def list(
        cls,
        project_id: str,
        featurelist_id: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List[SecondaryDatasetConfigurations]:
        """Returns list of secondary dataset configurations.

        .. versionadded:: v2.23

        Parameters
        ----------
        project_id: str
            The Id of project
        featurelist_id: Optional[str]
            Id of the feature list to filter the secondary datasets configurations

        Returns
        -------
        secondary_dataset_configurations : list of SecondaryDatasetConfigurations
            The requested list of secondary dataset configurations for a given project

        Examples
        --------
        .. code-block:: python

            pid = '5fd06afce2456ec1e9d20457'
            secondary_dataset_configs = dr.SecondaryDatasetConfigurations.list(pid)
            >>> secondary_dataset_configs[0]
                {
                     'created': datetime.datetime(2020, 12, 9, 6, 16, 22, tzinfo=tzutc()),
                     'creator_full_name': u'abc@datarobot.com',
                     'creator_user_id': u'asdf4af1gf4bdsd2fba1de0a',
                     'credential_ids': None,
                     'featurelist_id': None,
                     'id': u'5fd1e86c589238a4e635e93d',
                     'is_default': True,
                     'name': u'My config',
                     'project_id': u'5fd06afce2456ec1e9d20457',
                     'project_version': None,
                     'secondary_datasets': [
                            {
                                'snapshot_policy': u'latest',
                                'identifier': u'profile',
                                'catalog_version_id': u'5fd06b4af24c641b68e4d88f',
                                'catalog_id': u'5fd06b4af24c641b68e4d88e'
                            },
                            {
                                'snapshot_policy': u'dynamic',
                                'identifier': u'transaction',
                                'catalog_version_id': u'5fd1e86c589238a4e635e98e',
                                'catalog_id': u'5fd1e86c589238a4e635e98d'
                            }
                     ]
                }
        """
        kwargs: Dict[str, Union[str, int]] = {}
        if featurelist_id:
            kwargs["featurelistId"] = featurelist_id
        if limit:
            kwargs["limit"] = limit
        if offset:
            kwargs["offset"] = offset

        url = cls._base_url.format(project_id)
        r_data = cls._client.get(url, params=kwargs).json()
        return [SecondaryDatasetConfigurations.from_server_data(item) for item in r_data["data"]]



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import cast, Dict, List, Union

import trafaret as t

from datarobot.insights.base import BaseInsight


class Residuals(BaseInsight):
    """Class for Residuals calculations. Use the standard methods of BaseInsight to compute
    and retrieve: compute, create, list, get.

    Usage example:

        ```python
        >>> from datarobot.insights import Residuals
        >>> Residuals.list("672e32de69b0b676ced54d9c")
        [<datarobot.insights.residuals.Residuals object at 0x7fbce8305ae0>]
        >>> Residuals.compute("672e32de69b0b676ced54d9c", data_slice_id="677ae1249695103ba9feff97")
        <datarobot.models.status_check_job.StatusCheckJob object at 0x7fbcf4054b80>
        >>> Residuals.list("672e32de69b0b676ced54d9c")
        [<datarobot.insights.residuals.Residuals object at 0x7fbce8305870>,
        <datarobot.insights.residuals.Residuals object at 0x7fbce8305690>]
        >>> Residuals.get("672e32de69b0b676ced54d9c", data_slice_id="677ae1249695103ba9feff97")
        <datarobot.insights.residuals.Residuals object at 0x7fbce83057b0>
        >>> Residuals.get("672e32de69b0b676ced54d9c", data_slice_id="677ae1249695103ba9feff97").histogram
        [{'interval_start': -33.37288135593221, 'interval_end': -32.525000000000006, 'occurrences': 1}, ...]
        ```
    """

    HISTOGRAM_BIN = t.Dict(
        {
            t.Key("interval_start"): t.Float(),
            t.Key("interval_end"): t.Float(),
            t.Key("occurrences"): t.Int(),
        }
    ).ignore_extra("*")

    INSIGHT_NAME = "residuals"
    INSIGHT_DATA = {
        t.Key("histogram"): t.List(HISTOGRAM_BIN),
        t.Key("coefficient_of_determination"): t.Float(),
        t.Key("residual_mean"): t.Float(),
        t.Key("standard_deviation"): t.Float(),
        t.Key("data"): t.List(t.List(t.Float())),
    }

    @property
    def histogram(self) -> List[Dict[str, Union[int, float]]]:
        """Residuals histogram."""
        return cast(List[Dict[str, Union[int, float]]], self.data["histogram"])

    @property
    def coefficient_of_determination(self) -> float:
        """Coefficient of determination."""
        return cast(float, self.data["coefficient_of_determination"])

    @property
    def residual_mean(self) -> float:
        """Residual mean."""
        return cast(float, self.data["residual_mean"])

    @property
    def standard_deviation(self) -> float:
        """Standard deviation."""
        return cast(float, self.data["standard_deviation"])

    @property
    def chart_data(self) -> List[List[float]]:
        """The rows of Residuals chart data in [actual, predicted, residual, row number] form."""
        return cast(List[List[float]], self.data["data"])



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import cast, List, Optional, TYPE_CHECKING, Union

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject
from datarobot.models.model import DatetimeModel, Model

if TYPE_CHECKING:
    from datarobot.models.api_object import ServerDataDictType


class ModelRecommendation(APIObject):
    """A collection of information about a recommended model for a project.

    Attributes
    ----------
    project_id : str
        the id of the project the model belongs to
    model_id : str
        the id of the recommended model
    recommendation_type : str
        the type of model recommendation
    """

    _base_recommended_path_template = "projects/{}/recommendedModels/"
    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("recommendation_type"): String,
        }
    ).ignore_extra("*")

    def __init__(self, project_id: str, model_id: str, recommendation_type: str) -> None:
        self.project_id = project_id
        self.model_id = model_id
        self.recommendation_type = recommendation_type

    def __repr__(self) -> str:
        return "ModelRecommendation({}, {}, {})".format(
            self.project_id, self.model_id, self.recommendation_type
        )

    @classmethod
    def get(
        cls,
        project_id: str,
        recommendation_type: Optional[str] = None,
    ) -> Optional[ModelRecommendation]:
        """
        Retrieves the default or specified by recommendation_type recommendation.

        Parameters
        ----------
        project_id : str
            The project's id.
        recommendation_type : enums.RECOMMENDED_MODEL_TYPE
            The type of recommendation to get. If None, returns the default recommendation.

        Returns
        -------
        recommended_model : ModelRecommendation

        """
        if recommendation_type is None:
            url = cls._base_recommended_path_template.format(project_id) + "recommendedModel/"
            return cls.from_location(url)
        else:
            recommendations = cls.get_all(project_id)
            return cls.get_recommendation(recommendations, recommendation_type)

    @classmethod
    def get_all(cls, project_id: str) -> List[ModelRecommendation]:
        """
        Retrieves all of the current recommended models for the project.


        Parameters
        ----------
        project_id : str
            The project's id.

        Returns
        -------
        recommended_models : list of ModelRecommendation
        """
        url = cls._base_recommended_path_template.format(project_id)
        response = ModelRecommendation._server_data(url)
        return [
            ModelRecommendation.from_server_data(cast("ServerDataDictType", data))
            for data in response
        ]

    @classmethod
    def get_recommendation(
        cls, recommended_models: List[ModelRecommendation], recommendation_type: str
    ) -> Optional[ModelRecommendation]:
        """
        Returns the model in the given list with the requested type.


        Parameters
        ----------
        recommended_models : list of ModelRecommendation
        recommendation_type : enums.RECOMMENDED_MODEL_TYPE
            the type of model to extract from the recommended_models list

        Returns
        -------
        recommended_model : ModelRecommendation or None if no model with the requested type exists
        """

        return next(
            (
                model
                for model in recommended_models
                if model.recommendation_type == recommendation_type
            ),
            None,
        )

    def get_model(self) -> Union[DatetimeModel, Model]:
        """
        Returns the Model associated with this ModelRecommendation.

        Returns
        -------
        recommended_model : Model or DatetimeModel if the project is datetime-partitioned
        """
        return Model.get(self.project_id, self.model_id)



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

from mypy_extensions import TypedDict
import trafaret as t

from datarobot.enums import RegisteredModelDeploymentSortKey, RegisteredModelSortDirection
from datarobot.models.api_object import APIObject
from datarobot.models.model_registry.common import UserMetadata
from datarobot.models.model_registry.deployment import VersionAssociatedDeployment
from datarobot.utils.pagination import unpaginate


class Tag(TypedDict):
    name: str
    value: str


class TagWithId(Tag):
    id: str


class MlpkgFileContents(TypedDict):
    all_time_series_prediction_intervals: Optional[bool]


class ModelDescription(TypedDict):
    description: Optional[str]
    model_name: Optional[str]
    model_creator_id: Optional[str]
    model_creator_name: Optional[str]
    model_creator_email: Optional[str]
    model_created_at: Optional[str]
    location: Optional[str]
    build_environment_type: Optional[str]


class Dataset(TypedDict):  # pylint: disable=missing-class-docstring
    dataset_name: Optional[str]
    training_data_catalog_id: Optional[str]
    training_data_catalog_version_id: Optional[str]
    training_data_creator_id: Optional[str]
    training_data_creator_name: Optional[str]
    training_data_creator_email: Optional[str]
    training_data_created_at: Optional[str]
    training_data_size: Optional[int]
    holdout_dataset_name: Optional[str]
    holdout_data_catalog_id: Optional[str]
    holdout_data_catalog_version_id: Optional[str]
    holdout_data_creator_id: Optional[str]
    holdout_data_creator_name: Optional[str]
    holdout_data_creator_email: Optional[str]
    holdout_data_created_at: Optional[str]
    target_histogram_baseline: Optional[str]
    baseline_segmented_by: Optional[str]


class BiasAndFairness(TypedDict):
    protected_features: Optional[List[str]]
    preferable_target_value: Optional[Union[str, int, bool]]
    fairness_metrics_set: Optional[str]
    fairness_threshold: Optional[float]


class Timeseries(TypedDict):  # pylint: disable=missing-class-docstring
    datetime_column_name: Optional[str]
    forecast_distance_column_name: Optional[str]
    forecast_point_column_name: Optional[str]
    series_column_name: Optional[str]
    datetime_column_format: Optional[str]
    forecast_distances: Optional[List[int]]
    forecast_distances_time_unit: Optional[str]
    feature_derivation_window_start: Optional[int]
    feature_derivation_window_end: Optional[int]
    effective_feature_derivation_window_start: Optional[int]
    effective_feature_derivation_window_end: Optional[int]
    is_new_series_support: Optional[bool]
    is_cross_series: Optional[bool]
    is_traditional_time_series: Optional[bool]


class Target(TypedDict):
    name: str
    type: str
    class_names: Optional[List[str]]
    class_count: Optional[int]
    prediction_threshold: Optional[float]
    prediction_probabilities_column: Optional[str]


class ModelKind(TypedDict):
    is_time_series: bool
    is_multiseries: bool
    is_unsupervised_learning: bool
    is_anomaly_detection_model: bool
    is_feature_discovery: bool
    is_combined_model: bool
    is_decision_flow: bool
    is_unstructured: bool


class ImportMeta(TypedDict):
    creator_id: str
    creator_username: str
    creator_full_name: Optional[str]
    date_created: str
    original_file_name: Optional[str]
    contains_featurelists: Optional[bool]
    contains_fear_pipeline: Optional[bool]
    contains_leaderboard_meta: Optional[bool]
    contains_project_meta: Optional[bool]


class ExternalTarget(TypedDict):
    name: str
    type: str
    class_names: Optional[List[str]]
    prediction_threshold: Optional[float]
    prediction_probabilities_column: Optional[str]


class ExternalDatasets(TypedDict):
    training_data_catalog_id: Optional[str]
    holdout_data_catalog_id: Optional[str]


class ExternalGeospatialMonitoring(TypedDict):
    primary_location_column: Optional[str]


class GeospatialMonitoring(TypedDict):
    primary_location_column: Optional[str]
    location_columns: Optional[List[str]]


class ScoringCodeMeta(TypedDict):
    location: Optional[str]
    data_robot_prediction_version: Optional[str]


class UseCaseDetails(TypedDict):
    id: Optional[str]
    name: Optional[str]
    creator_name: Optional[str]
    creator_id: str
    creator_email: Optional[str]
    created_at: str


class CustomModelDetails(TypedDict):
    id: Optional[str]
    creator_name: Optional[str]
    creator_id: str
    creator_email: Optional[str]
    created_at: str
    version_label: Optional[str]


class SourceMeta(TypedDict):  # pylint: disable=missing-class-docstring
    project_id: Optional[str]
    project_name: Optional[str]
    project_creator_id: Optional[str]
    project_creator_name: Optional[str]
    project_creator_email: Optional[str]
    project_created_at: Optional[str]
    environment_url: Optional[str]
    scoring_code: Optional[ScoringCodeMeta]
    decision_flow_id: Optional[str]
    decision_flow_version_id: Optional[str]
    fips_140_2_enabled: Optional[bool]
    use_case_details: Optional[UseCaseDetails]
    custom_model_details: Optional[CustomModelDetails]


TRegisteredModelVersion = TypeVar("TRegisteredModelVersion", bound="RegisteredModelVersion")


class RegisteredModelVersion(APIObject):
    """
    Represents a version of a registered model.

    Parameters
    ----------
    id : str
        The ID of the registered model version.
    registered_model_id : str
        The ID of the parent registered model.
    registered_model_version : int
        The version of the registered model.
    name : str
        The name of the registered model version.
    model_id : str
        The ID of the model.
    model_execution_type : str
        Type of model package (version). `dedicated` (native DataRobot models) and
        custom_inference_model` (user added inference models) both execute on DataRobot
        prediction servers, `external` do not
    is_archived : bool
        Whether the model package (version) is permanently archived (cannot be used in deployment or
            replacement)
    import_meta : ImportMeta
        Information from when this Model Package (version) was first saved.
    source_meta : SourceMeta
        Meta information from where this model was generated
    model_kind : ModelKind
        Model attribute information.
    target : Target
        Target information for the registered model version.
    model_description : ModelDescription
        Model description information.
    datasets : Dataset
        Dataset information for the registered model version.
    timeseries : Timeseries
        Timeseries information for the registered model version.
    bias_and_fairness : BiasAndFairness
        Bias and fairness information for the registered model version.
    is_deprecated : bool
        Whether the model package (version) is deprecated (cannot be used in deployment or
            replacement)
    permissions : List[str]
        Permissions for the registered model version.
    active_deployment_count : int or None
        Number of the active deployments associated with the registered model version.
    build_status : str or None
        Model package (version) build status. One of `complete`, `inProgress`, `failed`.
    user_provided_id : str or None
        User provided ID for the registered model version.
    updated_at : str or None
        The time the registered model version was last updated.
    updated_by : UserMetadata or None
        The user who last updated the registered model version.
    tags : List[TagWithId] or None
        The tags associated with the registered model version.
    mlpkg_file_contents : str or None
        The contents of the model package file.
    """

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("registered_model_id"): t.String,
            t.Key("registered_model_version"): t.Int,
            t.Key("name"): t.String,
            t.Key("model_id"): t.String,
            t.Key("model_execution_type"): t.String,
            t.Key("is_archived"): t.Bool,
            t.Key("import_meta"): t.Dict().allow_extra("*"),
            t.Key("source_meta"): t.Dict().allow_extra("*"),
            t.Key("model_kind"): t.Dict().allow_extra("*"),
            t.Key("target"): t.Dict().allow_extra("*"),
            t.Key("model_description"): t.Dict().allow_extra("*"),
            t.Key("datasets"): t.Dict().allow_extra("*"),
            t.Key("timeseries"): t.Dict().allow_extra("*"),
            t.Key("is_deprecated"): t.Bool,
            t.Key("permissions"): t.List(t.String),
            t.Key("active_deployment_count"): t.Int,
            t.Key("bias_and_fairness", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
            t.Key("build_status", optional=True): t.Or(t.String, t.Null),
            t.Key("user_provided_id", optional=True): t.Or(t.String, t.Null),
            t.Key("updated_at", optional=True): t.Or(t.String, t.Null),
            t.Key("updated_by", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
            t.Key("tags", optional=True): t.Or(t.List(t.Dict().allow_extra("*")), t.Null),
            t.Key("mlpkg_file_contents", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
            t.Key("geospatial_monitoring", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        registered_model_id: str,
        registered_model_version: int,
        name: str,
        model_id: str,
        model_execution_type: str,
        is_archived: bool,
        import_meta: ImportMeta,
        source_meta: SourceMeta,
        model_kind: ModelKind,
        target: Target,
        model_description: ModelDescription,
        datasets: Dataset,
        timeseries: Timeseries,
        is_deprecated: bool,
        permissions: List[str],
        active_deployment_count: int,
        bias_and_fairness: Optional[BiasAndFairness] = None,
        build_status: Optional[str] = None,
        user_provided_id: Optional[str] = None,
        updated_at: Optional[str] = None,
        updated_by: Optional[UserMetadata] = None,
        tags: Optional[List[TagWithId]] = None,
        mlpkg_file_contents: Optional[MlpkgFileContents] = None,
        geospatial_monitoring: Optional[GeospatialMonitoring] = None,
    ):
        self.id = id
        self.registered_model_id = registered_model_id
        self.registered_model_version = registered_model_version
        self.name = name
        self.model_id = model_id
        self.model_execution_type = model_execution_type
        self.is_archived = is_archived
        self.import_meta = import_meta
        self.source_meta = source_meta
        self.model_kind = model_kind
        self.target = target
        self.model_description = model_description
        self.datasets = datasets
        self.timeseries = timeseries
        self.bias_and_fairness = bias_and_fairness
        self.build_status = build_status
        self.user_provided_id = user_provided_id
        self.updated_at = updated_at
        self.updated_by = updated_by
        self.tags = tags
        self.mlpkg_file_contents = mlpkg_file_contents
        self.is_deprecated = is_deprecated
        self.permissions = permissions
        self.active_deployment_count = active_deployment_count
        self.geo_spatial_monitoring = geospatial_monitoring

    @classmethod
    def create_for_leaderboard_item(
        cls: Type[TRegisteredModelVersion],
        model_id: str,
        name: Optional[str] = None,
        prediction_threshold: Optional[float] = None,
        distribution_prediction_model_id: Optional[str] = None,
        description: Optional[str] = None,
        compute_all_ts_intervals: Optional[bool] = None,
        registered_model_name: Optional[str] = None,
        registered_model_id: Optional[str] = None,
        tags: Optional[List[Tag]] = None,
        registered_model_tags: Optional[List[Tag]] = None,
        registered_model_description: Optional[str] = None,
    ) -> TRegisteredModelVersion:
        """

        Parameters
        ----------

        model_id : str
            ID of the DataRobot model.
        name : str or None
            Name of the version (model package).
        prediction_threshold : float or None
            Threshold used for binary classification in predictions.
        distribution_prediction_model_id : str or None
            ID of the DataRobot distribution prediction model
            trained on predictions from the DataRobot model.
        description : str or None
            Description of the version (model package).
        compute_all_ts_intervals : bool or None
            Whether to compute all time series prediction intervals (1-100 percentiles).
        registered_model_name : Optional[str]
            Name of the new registered model that will be created from this model package (version).
            The model package (version) will be created as version 1 of the created registered model.
            If neither registeredModelName nor registeredModelId is provided,
            it defaults to the model package (version) name. Mutually exclusive with registeredModelId.
        registered_model_id : Optional[str]
            Creates a model package (version) as a new version for the provided registered model ID.
            Mutually exclusive with registeredModelName.
        tags : Optional[List[Tag]]
            Tags for the registered model version.
        registered_model_tags: Optional[List[Tag]]
            Tags for the registered model.
        registered_model_description: Optional[str]
            Description for the registered model.

        Returns
        -------
        regitered_model_version : RegisteredModelVersion
            A new registered model version object.

        """
        url = "modelPackages/fromLeaderboard/"
        payload: Dict[str, Any] = {
            "modelId": model_id,
        }
        if name is not None:
            payload["name"] = name
        if prediction_threshold is not None:
            payload["predictionThreshold"] = prediction_threshold
        if distribution_prediction_model_id is not None:
            payload["distributionPredictionModelId"] = distribution_prediction_model_id
        if description is not None:
            payload["description"] = description
        if compute_all_ts_intervals is not None:
            payload["computeAllTsIntervals"] = compute_all_ts_intervals
        if registered_model_name is not None:
            payload["registeredModelName"] = registered_model_name
        if registered_model_id is not None:
            payload["registeredModelId"] = registered_model_id
        if tags is not None:
            payload["tags"] = tags
        if registered_model_tags is not None:
            payload["registeredModelTags"] = registered_model_tags
        if registered_model_description is not None:
            payload["registeredModelDescription"] = registered_model_description

        response = cls._client.post(url=url, json=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def create_for_external(
        cls: Type[TRegisteredModelVersion],
        name: str,
        target: ExternalTarget,
        model_id: Optional[str] = None,
        model_description: Optional[ModelDescription] = None,
        datasets: Optional[ExternalDatasets] = None,
        timeseries: Optional[Timeseries] = None,
        registered_model_name: Optional[str] = None,
        registered_model_id: Optional[str] = None,
        tags: Optional[List[Tag]] = None,
        registered_model_tags: Optional[List[Tag]] = None,
        registered_model_description: Optional[str] = None,
        geospatial_monitoring: Optional[ExternalGeospatialMonitoring] = None,
    ) -> TRegisteredModelVersion:
        """
        Create a new registered model version from an external model.

        Parameters
        ----------
        name : str
            Name of the registered model version.
        target : ExternalTarget
            Target information for the registered model version.
        model_id : Optional[str]
            Model ID of the registered model version.
        model_description : Optional[ModelDescription]
            Information about the model.
        datasets : Optional[ExternalDatasets]
            Dataset information for the registered model version.
        timeseries : Optional[Timeseries]
            Timeseries properties for the registered model version.
        registered_model_name : Optional[str]
            Name of the new registered model that will be created from this model package (version).
            The model package (version) will be created as version 1 of the created registered model.
            If neither registeredModelName nor registeredModelId is provided,
            it defaults to the model package (version) name. Mutually exclusive with registeredModelId.
        registered_model_id : Optional[str]
            Creates a model package (version) as a new version for the provided registered model ID.
            Mutually exclusive with registeredModelName.
        tags : Optional[List[Tag]]
            Tags for the registered model version.
        registered_model_tags: Optional[List[Tag]]
            Tags for the registered model.
        registered_model_description: Optional[str]
            Description for the registered model.
        geospatial_monitoring: Optional[ExternalGeospatialMonitoring]
            Geospatial monitoring settings for the registered model version.

        Returns
        -------
        registered_model_version : RegisteredModelVersion
            A new registered model version object.

        """
        url = "modelPackages/fromJSON/"
        payload: Dict[str, Any] = {
            "name": name,
            "target": target,
        }
        if model_id is not None:
            payload["modelId"] = model_id
        if model_description is not None:
            payload["modelDescription"] = model_description
        if datasets is not None:
            payload["datasets"] = datasets
        if timeseries is not None:
            payload["timeseries"] = timeseries
        if registered_model_name is not None:
            payload["registeredModelName"] = registered_model_name
        if registered_model_id is not None:
            payload["registeredModelId"] = registered_model_id
        if tags is not None:
            payload["tags"] = tags
        if registered_model_tags is not None:
            payload["registeredModelTags"] = registered_model_tags
        if registered_model_description is not None:
            payload["registeredModelDescription"] = registered_model_description
        if geospatial_monitoring is not None:
            payload["geospatialMonitoring"] = geospatial_monitoring

        response = cls._client.post(url=url, json=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def create_for_custom_model_version(
        cls: Type[TRegisteredModelVersion],
        custom_model_version_id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        registered_model_name: Optional[str] = None,
        registered_model_id: Optional[str] = None,
        tags: Optional[List[Tag]] = None,
        registered_model_tags: Optional[List[Tag]] = None,
        registered_model_description: Optional[str] = None,
    ) -> TRegisteredModelVersion:
        """
        Create a new registered model version from a custom model version.

        Parameters
        ----------
        custom_model_version_id : str
            ID of the custom model version.
        name : Optional[str]
            Name of the registered model version.
        description : Optional[str]
            Description of the registered model version.
        registered_model_name : Optional[str]
            Name of the new registered model that will be created from this model package (version).
            The model package (version) will be created as version 1 of the created registered model.
            If neither registeredModelName nor registeredModelId is provided,
            it defaults to the model package (version) name. Mutually exclusive with registeredModelId.
        registered_model_id : Optional[str]
            Creates a model package (version) as a new version for the provided registered model ID.
            Mutually exclusive with registeredModelName.
        tags : Optional[List[Tag]]
            Tags for the registered model version.
        registered_model_tags: Optional[List[Tag]]
            Tags for the registered model.
        registered_model_description: Optional[str]
            Description for the registered model.

        Returns
        -------
        registered_model_version : RegisteredModelVersion
            A new registered model version object.

        """
        url = "modelPackages/fromCustomModelVersion/"
        payload: Dict[str, Any] = {
            "customModelVersionId": custom_model_version_id,
        }
        if name is not None:
            payload["name"] = name
        if description is not None:
            payload["description"] = description
        if registered_model_name is not None:
            payload["registeredModelName"] = registered_model_name
        if registered_model_id is not None:
            payload["registeredModelId"] = registered_model_id
        if tags is not None:
            payload["tags"] = tags
        if registered_model_tags is not None:
            payload["registeredModelTags"] = registered_model_tags
        if registered_model_description is not None:
            payload["registeredModelDescription"] = registered_model_description

        response = cls._client.post(url=url, json=payload)
        return cls.from_server_data(response.json())

    def list_associated_deployments(
        self,
        search: Optional[str] = None,
        sort_key: Optional[RegisteredModelDeploymentSortKey] = None,
        sort_direction: Optional[RegisteredModelSortDirection] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List[VersionAssociatedDeployment]:
        """
        Retrieve a list of deployments associated with this registered model version.

        Parameters
        ----------
        search : Optional[str]
        sort_key : Optional[RegisteredModelDeploymentSortKey]
        sort_direction : Optional[RegisteredModelSortDirection]
        limit : Optional[int]
        offset : Optional[int]

        Returns
        -------
        deployments : List[VersionAssociatedDeployment]
            A list of deployments associated with this registered model version.

        """
        params: Dict[str, Any] = {}
        if sort_key:
            params["sortKey"] = sort_key.value
        if sort_direction:
            params["sortDirection"] = sort_direction.value
        if search:
            params["search"] = search
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        url = f"registeredModels/{self.registered_model_id}/versions/{self.id}/deployments/"
        data = unpaginate(url, params, self._client)
        return [VersionAssociatedDeployment.from_server_data(data_point) for data_point in data]



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from mypy_extensions import TypedDict
import trafaret as t

from datarobot.enums import enum_to_list, LLMTestConfigurationType, PromptSamplingStrategy
from datarobot.models.api_object import APIObject
from datarobot.models.genai.evaluation_dataset_configuration import (
    evaluation_dataset_configuration_trafaret,
    EvaluationDatasetConfiguration,
)
from datarobot.models.genai.insights_configuration import (
    insight_configuration_trafaret,
    InsightsConfiguration,
    InsightsConfigurationDict,
)
from datarobot.models.genai.playground import Playground
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.use_cases.utils import get_use_case_id, resolve_use_cases, UseCaseLike
from datarobot.utils.pagination import unpaginate


def get_entity_id(entity: Union[LLMTestConfiguration, Playground, UseCase, str]) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : ApiObject or str
        May be the entity ID or the entity name.

    Returns
    -------
    id : str
        Entity ID
    """
    if isinstance(entity, str):
        return entity

    return entity.id


class DatasetEvaluationDict(TypedDict):
    evaluation_name: str
    evaluation_dataset_configuration_id: Optional[str]
    evaluation_dataset_name: str
    ootb_dataset: Optional[OOTBDatasetDict]
    insight_configuration: InsightsConfigurationDict
    insight_grading_criteria: InsightGradingCriteriaDict
    max_num_prompts: Optional[int]
    prompt_sampling_strategy: Optional[PromptSamplingStrategy]


class DatasetEvaluationRequestDict(TypedDict):
    ootb_dataset_name: str
    evaluation_name: str
    evaluation_dataset_configuration_id: Optional[str]
    insight_configuration: InsightsConfigurationDict
    insight_grading_criteria: InsightGradingCriteriaDict
    max_num_prompts: Optional[int]
    prompt_sampling_strategy: Optional[PromptSamplingStrategy]


class OOTBDatasetDict(TypedDict):
    dataset_url: Optional[str]
    dataset_name: str
    prompt_column_name: str
    response_column_name: Optional[str]
    rows_count: int
    warning: Optional[str]


class InsightGradingCriteriaDict(TypedDict):
    pass_threshold: int


class LLMTestGradingCriteriaDict(TypedDict):
    pass_threshold: int


class DatasetIdentifierDict(TypedDict):
    dataset_name: str
    dataset_id: Optional[str]


class DatasetsCompatibilityDict(TypedDict):
    insight_name: str
    incompatible_datasets: List[DatasetIdentifierDict]


ootb_dataset_trafaret = t.Dict(
    {
        t.Key("dataset_url", optional=True): t.Or(t.String, t.Null),
        t.Key("dataset_name"): t.String,
        t.Key("prompt_column_name"): t.String,
        t.Key("response_column_name", optional=True): t.Or(t.String, t.Null),
        t.Key("rows_count"): t.Int,
        t.Key("warning", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


insight_grading_criteria_trafaret = t.Dict(
    {
        t.Key("pass_threshold"): t.Int,
    }
).ignore_extra("*")

llm_test_grading_criteria_trafaret = t.Dict(
    {
        t.Key("pass_threshold"): t.Int,
    }
).ignore_extra("*")


dataset_evaluation_trafaret = t.Dict(
    {
        t.Key("evaluation_name"): t.String(allow_blank=True),
        t.Key("evaluation_dataset_configuration_id", optional=True): t.Or(t.String, t.Null),
        t.Key("evaluation_dataset_name"): t.String(allow_blank=True),
        t.Key("ootb_dataset", optional=True): t.Or(ootb_dataset_trafaret, t.Null),
        t.Key("insight_configuration"): insight_configuration_trafaret,
        t.Key("insight_grading_criteria"): insight_grading_criteria_trafaret,
        t.Key("max_num_prompts", optional=True): t.Or(t.Int, t.Null),
        t.Key("prompt_sampling_strategy", optional=True): t.Or(
            t.Enum(*enum_to_list(PromptSamplingStrategy)), t.Null
        ),
    }
).ignore_extra("*")


llm_test_configuration_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String(allow_blank=True),
        t.Key("description"): t.String(allow_blank=True),
        t.Key("dataset_evaluations"): t.List(dataset_evaluation_trafaret),
        t.Key("use_case_id", optional=True): t.Or(t.String, t.Null),
        t.Key("creation_date", optional=True): t.Or(t.String, t.Null),
        t.Key("creation_user_id", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_test_grading_criteria"): llm_test_grading_criteria_trafaret,
        t.Key("is_out_of_the_box_test_configuration"): t.Bool,
        t.Key("warnings", optional=True): t.List(t.Dict().allow_extra("*")),
    }
).ignore_extra("*")

dataset_identifier_trafaret = t.Dict(
    {
        t.Key("dataset_name"): t.String(allow_blank=True),
        t.Key("dataset_id", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")

insight_to_eval_datasets_compatibility_trafaret = t.Dict(
    {
        t.Key("insight_name"): t.String,
        t.Key("incompatible_datasets"): t.List(dataset_identifier_trafaret),
    }
).ignore_extra("*")

llm_test_configuration_supported_insights_trafaret = t.Dict(
    {
        t.Key("supported_insight_configurations"): t.List(insight_configuration_trafaret),
        t.Key("datasets_compatibility"): t.List(insight_to_eval_datasets_compatibility_trafaret),
    }
).ignore_extra("*")


class DatasetIdentifier(APIObject):
    """
    Metadata for a DataRobot GenAI dataset identifier.

    Attributes
    ----------
    dataset_name: str
        The name of the dataset.
    dataset_id: str or None, optional
        The ID of the dataset.
    """

    _converter = dataset_identifier_trafaret

    def __init__(
        self,
        dataset_name: str,
        dataset_id: Optional[str] = None,
    ):
        self.dataset_name = dataset_name
        self.dataset_id = dataset_id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(dataset_name={self.dataset_name})"

    def to_dict(self) -> DatasetIdentifierDict:
        return {
            "dataset_name": self.dataset_name,
            "dataset_id": self.dataset_id,
        }


class DatasetsCompatibility(APIObject):
    """
    Metadata for a DataRobot GenAI LLM test configuration supported insights datasets compatibility.

    Attributes
    ----------
    insight_name: str
        The name of the insight.
    incompatible_datasets: list[DatasetIdentifier]
        The incompatible datasets for the insight.
    """

    _converter = insight_to_eval_datasets_compatibility_trafaret

    def __init__(
        self,
        insight_name: str,
        incompatible_datasets: List[Dict[str, Any]],
    ):
        self.insight_name = insight_name
        self.incompatible_datasets = [
            DatasetIdentifier.from_server_data(dataset_identifier)
            for dataset_identifier in incompatible_datasets
        ]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(insight_name={self.insight_name})"

    def to_dict(self) -> DatasetsCompatibilityDict:
        return {
            "insight_name": self.insight_name,
            "incompatible_datasets": [dataset.to_dict() for dataset in self.incompatible_datasets],
        }


class NonOOTBDataset(APIObject):
    """
    Metadata for a DataRobot GenAI non out-of-the-box (OOTB) LLM compliance test dataset.
    """

    _path = "api/v2/genai/llmTestConfigurations/nonOotbDatasets"

    _converter = evaluation_dataset_configuration_trafaret

    def __init__(
        self,
        **kwargs: Any,
    ):
        self.non_ootb_dataset = EvaluationDatasetConfiguration(**kwargs)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.non_ootb_dataset.id}, "
            f"dataset_id={self.non_ootb_dataset.dataset_id})"
        )

    @classmethod
    def list(
        cls,
        use_case: Optional[Union[UseCase, str]] = None,
    ) -> List[NonOOTBDataset]:
        """
        List all non out-of-the-box datasets available to the user.

        Returns
        -------
        non_ootb_datasets : list[NonOOTBDataset]
            Returns a list of non out-of-the-box datasets.
        """
        url = f"{cls._client.domain}/{cls._path}/"
        params = resolve_use_cases(use_cases=use_case, params={}, use_case_key="use_case_id")
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]


class OOTBDataset(APIObject):
    """
    Metadata for a DataRobot GenAI out-of-the-box LLM compliance test dataset.

    Attributes
    ----------
    dataset_name: str
        The name of the dataset.
    prompt_column_name : str
        The name of the prompt column.
    response_column_name: str or None, optional
        The name of the response column, if any.
    dataset_url: str or None, optional
        The URL of the dataset.
    rows_count: int
        The number of rows in the dataset.
    warning: str or None, optional
        A warning message regarding the contents of the dataset, if any.
    """

    _path = "api/v2/genai/llmTestConfigurations/ootbDatasets"

    _converter = ootb_dataset_trafaret

    def __init__(
        self,
        dataset_name: str,
        prompt_column_name: str,
        rows_count: int,
        dataset_url: Optional[str] = None,
        response_column_name: Optional[str] = None,
        warning: Optional[str] = None,
    ):
        self.dataset_url = dataset_url
        self.dataset_name = dataset_name
        self.prompt_column_name = prompt_column_name
        self.response_column_name = response_column_name
        self.rows_count = rows_count
        self.warning = warning

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(dataset_url={self.dataset_url})"

    def to_dict(self) -> OOTBDatasetDict:
        return {
            "dataset_url": self.dataset_url,
            "dataset_name": self.dataset_name,
            "prompt_column_name": self.prompt_column_name,
            "response_column_name": self.response_column_name,
            "rows_count": self.rows_count,
            "warning": self.warning,
        }

    @classmethod
    def list(
        cls,
    ) -> List[OOTBDataset]:
        """
        List all out-of-the-box datasets available to the user.

        Returns
        -------
        ootb_datasets : list[OOTBDataset]
            Returns a list of out-of-the-box datasets.
        """
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, None, cls._client)
        return [cls.from_server_data(data) for data in r_data]


class LLMTestGradingCriteria(APIObject):
    """
    Metadata for a GenAI LLM test grading criteria.

    Attributes
    ----------
    pass_threshold: int
        The threshold to pass the test.
    """

    _converter = insight_grading_criteria_trafaret

    def __init__(
        self,
        pass_threshold: int,
    ):
        self.pass_threshold = pass_threshold

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(pass_threshold={self.pass_threshold})"

    def to_dict(self) -> LLMTestGradingCriteriaDict:
        return {
            "pass_threshold": self.pass_threshold,
        }


class InsightGradingCriteria(APIObject):
    """
    Metadata for a GenAI LLM test insight grading criteria.

    Attributes
    ----------
    pass_threshold: int
        The threshold to pass the test.
    """

    _converter = insight_grading_criteria_trafaret

    def __init__(
        self,
        pass_threshold: int,
    ):
        self.pass_threshold = pass_threshold

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(pass_threshold={self.pass_threshold})"

    def to_dict(self) -> InsightGradingCriteriaDict:
        return {
            "pass_threshold": self.pass_threshold,
        }


class DatasetEvaluation(APIObject):
    """
    Metadata for a DataRobot GenAI dataset evaluation.

    Attributes
    ----------
    evaluation_name: str
        The name of the evaluation.
    evaluation_dataset_configuration_id: str or None, optional
        The ID of the evaluation dataset configuration for custom datasets.
    evaluation_dataset_name: str
        The name of the evaluation dataset.
    ootb_dataset: OOTBDataset or None, optional
        Out-of-the-box dataset.
    insight_configuration: InsightsConfiguration
        The insight to calculate for this dataset.
    insight_grading_criteria: InsightGradingCriteria
        The criteria to use for grading the results.
    max_num_prompts: int
        The maximum number of prompts to use for the evaluation.
    prompt_sampling_strategy: PromptSamplingStrategy
        The prompt sampling strategy for the dataset evaluation.
    """

    _converter = dataset_evaluation_trafaret

    def __init__(
        self,
        evaluation_name: str,
        insight_configuration: Dict[str, Any],
        insight_grading_criteria: Dict[str, Any],
        evaluation_dataset_name: str,
        max_num_prompts: Optional[int] = None,
        prompt_sampling_strategy: Optional[PromptSamplingStrategy] = None,
        evaluation_dataset_configuration_id: Optional[str] = None,
        ootb_dataset: Optional[Dict[str, Any]] = None,
    ):
        self.evaluation_name = evaluation_name
        self.evaluation_dataset_name = evaluation_dataset_name
        self.evaluation_dataset_configuration_id = evaluation_dataset_configuration_id
        self.ootb_dataset = OOTBDataset.from_server_data(ootb_dataset) if ootb_dataset else None
        self.insight_configuration = InsightsConfiguration.from_server_data(insight_configuration)
        self.insight_grading_criteria = InsightGradingCriteria.from_server_data(
            insight_grading_criteria
        )
        self.max_num_prompts = max_num_prompts
        self.prompt_sampling_strategy = prompt_sampling_strategy

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(evaluation_name={self.evaluation_name})"

    def to_dict(self) -> DatasetEvaluationDict:
        return {
            "evaluation_name": self.evaluation_name,
            "evaluation_dataset_configuration_id": self.evaluation_dataset_configuration_id,
            "evaluation_dataset_name": self.evaluation_dataset_name,
            "ootb_dataset": self.ootb_dataset.to_dict() if self.ootb_dataset else None,
            "insight_configuration": self.insight_configuration.to_dict(),
            "insight_grading_criteria": self.insight_grading_criteria.to_dict(),
            "max_num_prompts": self.max_num_prompts,
            "prompt_sampling_strategy": self.prompt_sampling_strategy,
        }


class LLMTestConfigurationSupportedInsights(APIObject):
    """
    Metadata for a DataRobot GenAI LLM test configuration supported insights.

    Attributes
    ----------
    supported_insight_configurations : list[InsightsConfiguration]
        The supported insights for LLM test configurations.
    """

    _path = "api/v2/genai/llmTestConfigurations/supportedInsights"

    _converter = llm_test_configuration_supported_insights_trafaret

    def __init__(
        self,
        supported_insight_configurations: List[Dict[str, Any]],
        datasets_compatibility: List[Dict[str, Any]],
    ):
        self.supported_insight_configurations = [
            InsightsConfiguration.from_server_data(insight_configuration)
            for insight_configuration in supported_insight_configurations
        ]
        self.datasets_compatibility = [
            DatasetsCompatibility.from_server_data(dataset_compatibility)
            for dataset_compatibility in datasets_compatibility
        ]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(supported_insight_configurations={self.supported_insight_configurations})"

    @classmethod
    def list(
        cls,
        use_case: Optional[Union[UseCase, str]] = None,
        playground: Optional[Union[Playground, str]] = None,
    ) -> LLMTestConfigurationSupportedInsights:
        """
        List all supported insights for a LLM test configuration.

        Parameters
        ----------
        use_case : Optional[Union[UseCase, str]], optional
            Returns only those supported insight configurations
            associated with a particular Use Case, specified by
            either the Use Case name or ID.
        playground : Optional[Union[Playground, str]], optional
            Returns only those supported insight configurations
            associated with a particular playground, specified by
            either the Playground or ID.

        Returns
        -------
        llm_test_configuration_supported_insights : LLMTestConfigurationSupportedInsights
            Returns the supported insight configurations for the
            LLM test configuration.
        """
        params = {
            "playground_id": get_entity_id(playground) if playground else None,
            "use_case_id": get_entity_id(use_case) if use_case else None,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.get(url, params=params)
        return cls.from_server_data(r_data.json())


class LLMTestConfiguration(APIObject):
    """
    Metadata for a DataRobot GenAI LLM test configuration.

    Attributes
    ----------
    id : str
        The LLM test configuration ID.
    name : str
        The LLM test configuration name.
    description : str
        The LLM test configuration description.
    dataset_evaluations : list[DatasetEvaluation]
        The dataset/insight combinations that make up the LLM test configuration.
    llm_test_grading_criteria : LLMTestGradingCriteria
        The criteria used to grade the result of the LLM test configuration.
    is_out_of_the_box_test_configuration : bool
        Whether this is an out-of-the-box configuration.
    use_case_id : Optional[str]
        The ID of the linked Use Case, if any.
    creation_date : Optional[str]
        The date the LLM test configuration was created, if any.
    creation_user_id : Optional[str]
        The ID of the creating user, if any.
    warnings: Optional[list[Dict[str, str]]]
        The warnings for the LLM test configuration, if any.
    """

    _path = "api/v2/genai/llmTestConfigurations"

    _converter = llm_test_configuration_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        dataset_evaluations: List[Dict[str, Any]],
        llm_test_grading_criteria: Dict[str, Any],
        is_out_of_the_box_test_configuration: bool,
        use_case_id: Optional[str] = None,
        creation_date: Optional[str] = None,
        creation_user_id: Optional[str] = None,
        warnings: Optional[List[Dict[str, str]] | None] = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.use_case_id = use_case_id
        self.dataset_evaluations = [
            DatasetEvaluation.from_server_data(dataset_evaluation)
            for dataset_evaluation in dataset_evaluations
        ]
        self.llm_test_grading_criteria = LLMTestGradingCriteria.from_server_data(
            llm_test_grading_criteria
        )
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.is_out_of_the_box_test_configuration = is_out_of_the_box_test_configuration
        self.warnings = warnings or []

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name={self.name})"

    @classmethod
    def create(
        cls,
        name: str,
        dataset_evaluations: List[DatasetEvaluationRequestDict],
        llm_test_grading_criteria: LLMTestGradingCriteria,
        use_case: Optional[Union[UseCase, str]] = None,
        description: Optional[str] = None,
    ) -> LLMTestConfiguration:
        """
        Creates a new LLM test configuration.

        Parameters
        ----------
        name : str
            The LLM test configuration name.
        dataset_evaluations : list[DatasetEvaluationRequestDict]
            The LLM test dataset evaluation requests.
        llm_test_grading_criteria : LLMTestGradingCriteria
            The LLM test grading criteria.
        use_case : Optional[Union[UseCase, str]], optional
            Use case to link to the created llm test configuration.
        description : Optional[str]
            The LLM test configuration description. If None, the default,
            description returns an empty string.

        Returns
        -------
        llm_test_configuration : LLMTestConfiguration
            The created LLM test configuration.
        """
        payload: Dict[str, Any] = {
            "name": name,
            "use_case_id": get_use_case_id(use_case, is_required=True),
            "dataset_evaluations": dataset_evaluations,
            "llm_test_grading_criteria": llm_test_grading_criteria.to_dict(),
        }
        if description:
            payload["description"] = description

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, json=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(
        cls,
        llm_test_configuration: Union[LLMTestConfiguration, str],
    ) -> LLMTestConfiguration:
        """
        Retrieve a single LLM Test configuration.

        Parameters
        ----------
        llm_test_configuration : LLMTestConfiguration or str
            The LLM test configuration to retrieve, either LLMTestConfiguration or LLMTestConfiguration ID.

        Returns
        -------
        llm_test_configuration : LLMTestConfiguration
            The requested LLM Test configuration.
        """
        url = f"{cls._client.domain}/{cls._path}/{get_entity_id(llm_test_configuration)}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        use_case: Optional[UseCaseLike] = None,
        test_config_type: Optional[LLMTestConfigurationType] = None,
    ) -> List[LLMTestConfiguration]:
        """
        List all LLM test configurations available to the user. If a Use Case is specified,
        results are restricted to only those configurations associated with that Use Case.

        Parameters
        ----------
        use_case : Optional[UseCaseLike], optional
            Returns only those configurations associated with a particular Use Case,
            specified by either the Use Case name or ID.
        test_config_type : Optional[LLMTestConfigurationType], optional
            Returns only configurations of the specified type. If not specified,
            the custom test configurations are returned.

        Returns
        -------
        llm_test_configurations : list[LLMTestConfiguration]
            Returns a list of LLM test configurations.
        """
        params = {}
        if test_config_type:
            params["test_config_type"] = test_config_type
        params = resolve_use_cases(use_cases=use_case, params=params, use_case_key="use_case_id")
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        dataset_evaluations: Optional[List[DatasetEvaluationRequestDict]] = None,
        llm_test_grading_criteria: Optional[LLMTestGradingCriteria] = None,
    ) -> LLMTestConfiguration:
        """
        Update the LLM test configuration.

        Parameters
        ----------
        name : Optional[str]
            The new LLM test configuration name.
        description : Optional[str]
            The new LLM test configuration description.
        dataset_evaluations : list[DatasetEvaluationRequestDict], optional
            The new dataset evaluation requests.
        llm_test_grading_criteria : LLMTestGradingCriteria, optional
            The new grading criteria.

        Returns
        -------
        llm_test_configuration : LLMTestConfiguration
            The updated LLM test configuration.
        """
        payload: Dict[str, Any] = {}
        if name is not None:
            payload["name"] = name
        if description is not None:
            payload["description"] = description
        if dataset_evaluations is not None:
            payload["dataset_evaluations"] = dataset_evaluations
        if llm_test_grading_criteria is not None:
            payload["llm_test_grading_criteria"] = llm_test_grading_criteria.to_dict()
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        r_data = self._client.patch(url, json=payload)
        return self.from_server_data(r_data.json())

    def delete(self) -> None:
        """
        Delete a single LLM test configuration.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, Iterable, List, Mapping, Optional, TYPE_CHECKING

import numpy as np
import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject

if TYPE_CHECKING:
    from datarobot.models.types import RocCurveEstimatedMetric


class RocCurveThresholdMixin:  # pylint: disable=missing-class-docstring
    roc_points: Optional[List[RocCurveEstimatedMetric]] = None

    @staticmethod
    def _validate_threshold(threshold: float) -> None:
        if threshold > 1 or threshold < 0:
            raise ValueError("threshold must be from [0, 1] interval")

    def estimate_threshold(self, threshold: float) -> RocCurveEstimatedMetric:
        """Return metrics estimation for given threshold.

        Parameters
        ----------
        threshold : float from [0, 1] interval
            Threshold we want estimation for

        Returns
        -------
        dict
            Dictionary of estimated metrics in form of {metric_name: metric_value}.
            Metrics are 'accuracy', 'f1_score', 'false_negative_score', 'true_negative_score',
            'true_negative_rate', 'matthews_correlation_coefficient', 'true_positive_score',
            'positive_predictive_value', 'false_positive_score', 'false_positive_rate',
            'negative_predictive_value', 'true_positive_rate'.

        Raises
        ------
        ValueError
            Given threshold isn't from [0, 1] interval
        """
        if self.roc_points is None:
            raise ValueError("ROC points must be set to estimate threshold.")
        self._validate_threshold(threshold)
        for roc_point in self.roc_points:
            if np.isclose(roc_point["threshold"], threshold):
                estimated_roc_point = roc_point
                break
        else:
            # if no exact match - pick closest ROC point with bigger threshold
            roc_points_with_bigger_threshold = [
                roc_point for roc_point in self.roc_points if roc_point["threshold"] > threshold
            ]
            estimated_roc_point = sorted(
                roc_points_with_bigger_threshold, key=lambda rp: rp["threshold"]
            )[0]
        return estimated_roc_point

    def get_best_f1_threshold(self) -> float:
        """Return value of threshold that corresponds to max F1 score.
        This is threshold that will be preselected in DataRobot when you open "ROC curve" tab.

        Returns
        -------
        float
            Threhold with best F1 score.
        """
        if self.roc_points is None:
            raise ValueError("ROC points must be set to get best f1 threshold.")
        roc_point_with_best_f1 = max(self.roc_points, key=lambda roc_point: roc_point["f1_score"])
        return roc_point_with_best_f1["threshold"]


RocPointsTrafaret = t.Dict(
    {
        t.Key("negative_class_predictions"): t.List(t.Float),
        t.Key("positive_class_predictions"): t.List(t.Float),
        t.Key("roc_points"): t.List(
            t.Dict(
                {
                    t.Key("accuracy"): t.Float,
                    t.Key("f1_score"): t.Float,
                    t.Key("false_negative_score"): Int,
                    t.Key("true_negative_score"): Int,
                    t.Key("true_positive_score"): Int,
                    t.Key("false_positive_score"): Int,
                    t.Key("true_negative_rate"): t.Float,
                    t.Key("false_positive_rate"): t.Float,
                    t.Key("true_positive_rate"): t.Float,
                    t.Key("matthews_correlation_coefficient"): t.Float,
                    t.Key("positive_predictive_value"): t.Float,
                    t.Key("negative_predictive_value"): t.Float,
                    t.Key("threshold"): t.Float,
                    t.Key("fraction_predicted_as_positive"): t.Float,
                    t.Key("fraction_predicted_as_negative"): t.Float,
                    t.Key("lift_positive"): t.Float,
                    t.Key("lift_negative"): t.Float,
                }
            ).ignore_extra("*")
        ),
    }
)

RocCurveTrafaret = (
    t.Dict(
        {
            t.Key("source"): String,
            t.Key("source_model_id"): String,
            t.Key("data_slice_id", optional=True): t.Or(String, t.Null),
        }
    )
    .merge(RocPointsTrafaret)
    .ignore_extra("*")
)


class RocCurve(APIObject, RocCurveThresholdMixin):
    """ROC curve data for model.

    Attributes
    ----------
    source : str
        ROC curve data source. Can be 'validation', 'crossValidation' or 'holdout'.
    roc_points : list of dict
        List of precalculated metrics associated with thresholds for ROC curve.
    negative_class_predictions : list of float
        List of predictions from example for negative class
    positive_class_predictions : list of float
        List of predictions from example for positive class
    source_model_id : str
        ID of the model this ROC curve represents; in some cases,
        insights from the parent of a frozen model may be used
    data_slice_id: str
        ID of the data slice this ROC curve represents.
    """

    _converter = RocCurveTrafaret

    def __init__(
        self,
        source: str,
        roc_points: List[RocCurveEstimatedMetric],
        negative_class_predictions: List[float],
        positive_class_predictions: List[float],
        source_model_id: str,
        data_slice_id: Optional[str] = None,
    ) -> None:
        self.source = source
        self.roc_points = roc_points
        self.negative_class_predictions = negative_class_predictions
        self.positive_class_predictions = positive_class_predictions
        self.source_model_id = source_model_id
        self.data_slice_id = data_slice_id

    def __repr__(self) -> str:
        return f"RocCurve({self.source})"

    @classmethod
    # type: ignore[override]
    def from_server_data(
        cls,
        data: Dict[str, Any],
        keep_attrs: Optional[Iterable[str]] = None,
        use_insights_format: bool = False,
        **kwargs: Mapping[str, Any],
    ) -> RocCurve:
        """
        Overwrite APIObject.from_server_data to handle roc curve data retrieved
        from either legacy URL or /insights/ new URL.

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place.
        keep_attrs : iterable
            List, set or tuple of the dotted namespace notations for attributes to keep within the
            object structure even if their values are None
        use_insights_format : Optional[bool]
            Whether to repack the data from the format used in the GET /insights/RocCur/ URL
            to the format used in the legacy URL.
        """
        if use_insights_format:
            data = cls._repack_insights_response(data)

        return super().from_server_data(data=data, keep_attrs=keep_attrs)

    @staticmethod
    def _repack_insights_response(server_data: Dict[str, Any]) -> Dict[str, Any]:
        """Repack the JSON sent by the GET /insights/ endpoint to match the format expected by the
        insight APIObject class.

        Parameters
        ----------
        server_data : dict
        {
            "id": "6474726c7ca961a4ebce068d",
            "entityId": "64747242956c7390bb15b206",
            "projectId": "647471b6b5d9cbd454f2cf63",
            "source": "validation",
            "dataSliceId": "647471b6b5d9cbd454f2ab99]
            "data": {'rocPoints': {}, 'positiveClassPredictions': {} 'negativeClassPredictions': {}}
        }

        Returns
        -------
        dict
        {
            "source": "validation",
            "rocPoints": {}.
            "positiveClassPredictions": {},
            "negativeClassPredictions": {},
            "sourceModelId": "64747242956c7390bb15b206",
            "dataSliceId": "647471b6b5d9cbd454f2ab99",
        }
        """
        return {
            "source": server_data["source"],
            "rocPoints": server_data["data"]["rocPoints"],
            "positiveClassPredictions": server_data["data"]["positiveClassPredictions"],
            "negativeClassPredictions": server_data["data"]["negativeClassPredictions"],
            "sourceModelId": server_data["entityId"],
            "dataSliceId": server_data["dataSliceId"],
        }


class SlicedRocCurve(RocCurve):
    """Wrapper around RocCurve to override `from_server_data` method"""

    @classmethod
    # type: ignore[override]
    def from_server_data(
        cls,
        data: Dict[str, Any],
        keep_attrs: Optional[Iterable[str]] = None,
        use_insights_format: bool = True,
        **kwargs: Mapping[str, Any],
    ) -> "RocCurve":
        """
        Overwrite RocCurve.from_server_data to set `use_insights_format=True` by default
        This is necessary for the correct transformation of data received from /insights endpoints

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place.
        keep_attrs : iterable
            List, set or tuple of the dotted namespace notations for attributes to keep within the
            object structure even if their values are None
        use_insights_format : Optional[bool]
            Whether to repack the data from the format used in the GET /insights/liftChart/ URL
            to the format used in the legacy URL.
        """
        if data.get("count"):
            # it's a list
            data = data["data"][0]

        return super().from_server_data(
            data=data, keep_attrs=keep_attrs, use_insights_format=use_insights_format, **kwargs
        )


class LabelwiseRocCurve(RocCurve):
    """Labelwise ROC curve data for one label and one source.

    Attributes
    ----------
    source : str
        ROC curve data source. Can be 'validation', 'crossValidation' or 'holdout'.
    roc_points : list of dict
        List of precalculated metrics associated with thresholds for ROC curve.
    negative_class_predictions : list of float
        List of predictions from example for negative class
    positive_class_predictions : list of float
        List of predictions from example for positive class
    source_model_id : str
        ID of the model this ROC curve represents; in some cases,
        insights from the parent of a frozen model may be used
    label : str
        Label name for
    kolmogorov_smirnov_metric : float
        Kolmogorov-Smirnov metric value for label
    auc : float
        AUC metric value for label
    """

    _converter = (
        t.Dict(
            {
                t.Key("label"): String,
                t.Key("kolmogorov_smirnov_metric"): t.Float,
                t.Key("auc"): t.Float,
            }
        )
        .merge(RocCurveTrafaret)
        .ignore_extra("*")
    )

    def __init__(
        self,
        source: str,
        roc_points: List[RocCurveEstimatedMetric],
        negative_class_predictions: List[float],
        positive_class_predictions: List[float],
        source_model_id: str,
        label: str,
        kolmogorov_smirnov_metric: float,
        auc: float,
    ) -> None:
        super().__init__(
            source=source,
            roc_points=roc_points,
            negative_class_predictions=negative_class_predictions,
            positive_class_predictions=positive_class_predictions,
            source_model_id=source_model_id,
        )
        self.label = label
        self.kolmogorov_smirnov_metric = kolmogorov_smirnov_metric
        self.auc = auc



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, List, Optional, Tuple, TYPE_CHECKING

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class FeatureAssociationMatrixDetailsDict(TypedDict):
        chart_type: Optional[str]
        values: Optional[List[Tuple[Any, Any, float]]]
        features: Optional[List[str]]
        types: Optional[List[str]]


class FeatureAssociationMatrixDetails(APIObject):
    """
    Plotting details for a pair of passed features present in the feature association matrix.

    Notes
    -----
    Projects created prior to v2.17 are not supported by this feature.

    Attributes
    ----------
    project_id : str
        Id of the project that contains the requested associations.
    chart_type : str
        Which type of plotting the pair of features gets in the UI.
        e.g. 'HORIZONTAL_BOX', 'VERTICAL_BOX', 'SCATTER' or 'CONTINGENCY'
    values : list
        The data triplets for pairwise plotting e.g.
        {"values": [[460.0, 428.5, 0.001], [1679.3, 259.0, 0.001], ...]
        The first entry of each list is a value of feature1, the second entry of each list is a
        value of feature2, and the third is the relative frequency of the pair of datapoints in the
        sample.
    features : list
        A list of the requested features, [feature1, feature2]
    types : list
        The type of `feature1` and `feature2`. Possible values: "CATEGORICAL", "NUMERIC"
    featurelist_id : str
        Id of the feature list to lookup FAM details for.
    """

    _path = "projects/{}/featureAssociationMatrixDetails/"
    _converter = t.Dict(
        {
            t.Key("chart_type"): String(),
            t.Key("values"): t.List(t.Tuple(t.Any(), t.Any(), t.Float())),
            t.Key("features"): t.List(String()),
            t.Key("types"): t.List(String()),
        }
    )

    def __init__(
        self,
        project_id: Optional[str] = None,
        chart_type: Optional[str] = None,
        values: Optional[List[Tuple[Any, Any, float]]] = None,
        features: Optional[List[str]] = None,
        types: Optional[List[str]] = None,
        featurelist_id: Optional[str] = None,
    ):
        self.project_id = project_id
        self.chart_type = chart_type
        self.values = values
        self.features = features
        self.types = types
        self.featurelist_id = featurelist_id

    def __repr__(self) -> str:
        return "{}(project_id={}, chart_type={}, values={}, features={}, types={})".format(
            self.__class__.__name__,
            self.project_id,
            self.chart_type,
            self.values,
            self.features,
            self.types,
        )

    @classmethod
    def get(
        cls,
        project_id: str,
        feature1: str,
        feature2: str,
        featurelist_id: Optional[str] = None,
    ) -> FeatureAssociationMatrixDetails:
        """
        Get a sample of the actual values used to measure the association between a pair of features

        .. versionadded:: v2.17

        Parameters
        ----------
        project_id : str
            Id of the project of interest.
        feature1 : str
            Feature name for the first feature of interest.
        feature2 : str
            Feature name for the second feature of interest.
        featurelist_id : str
            Optional, the feature list to lookup FAM data for. By default, depending on the type of
            the project "Informative Features" or "Timeseries Informative Features" list will be
            used.

        Returns
        -------
        FeatureAssociationMatrixDetails
            The feature association plotting for provided pair of features.
        """
        url = cls._path.format(project_id)
        params = {"feature1": feature1, "feature2": feature2}
        if featurelist_id:
            params["featurelistId"] = featurelist_id
        response = cls._client.get(url, params=params)
        fam_details = cls.from_server_data(response.json())
        fam_details.project_id = project_id
        fam_details.featurelist_id = featurelist_id
        return fam_details

    def to_dict(self) -> FeatureAssociationMatrixDetailsDict:
        return {
            "chart_type": self.chart_type,
            "values": self.values,
            "features": self.features,
            "types": self.types,
        }



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, Optional

import trafaret as t
from trafaret import String

from datarobot.enums import (
    ModerationGuardAction,
    ModerationGuardConditionLogic,
    ModerationGuardConditionOperator,
)
from datarobot.models.api_object import APIObject


class GuardInterventionCondition(APIObject):
    """Defines a condition for intervention."""

    _converter = t.Dict(
        {
            t.Key("comparator"): t.Enum(*[e.value for e in ModerationGuardConditionOperator]),
            t.Key("comparand"): t.Or(
                t.Float(),
                t.String(),
                t.Bool(),
                t.List(
                    t.String(),
                ),
            ),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self: GuardInterventionCondition, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "{}(comparator={!r}, comparand={!r})".format(
            self.__class__.__name__,
            self.comparator,
            self.comparand,
        )

    def _set_values(
        self: GuardInterventionCondition,
        comparator: ModerationGuardConditionOperator,
        comparand: Any,
    ) -> None:
        self.comparator = ModerationGuardConditionOperator(comparator)
        self.comparand = comparand

    def to_dict(self) -> Dict[str, str]:
        return {
            "comparator": self.comparator.value,
            "comparand": self.comparand,
        }


class GuardInterventionForTemplate(APIObject):
    """
    Defines the intervention conditions and actions a guard can take.
    Configuration schema differs slightly from template because changes were requested after
    templates were baked in.
    """

    _converter = t.Dict(
        {
            t.Key("action"): t.Enum(*[e.value for e in ModerationGuardAction]),
            t.Key("conditions"): t.List(GuardInterventionCondition.schema),
            t.Key("modify_message"): t.String(allow_blank=True),
            t.Key("allowed_actions"): t.List(t.Enum(*[e.value for e in ModerationGuardAction])),
            t.Key("condition_logic", optional=True): t.Enum(
                *[e.value for e in ModerationGuardConditionLogic]
            ),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self: GuardInterventionForTemplate, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "{}(action={!r})".format(
            self.__class__.__name__,
            self.action,
        )

    def _set_values(
        self: GuardInterventionForTemplate,
        action: ModerationGuardAction | str,
        conditions: list[GuardInterventionCondition | str],
        modify_message: str,
        allowed_actions: list[ModerationGuardAction | str],
        condition_logic: Optional[
            ModerationGuardConditionLogic
        ] = ModerationGuardConditionLogic.ANY,
    ) -> None:
        """
        Initialize an intervention object based on values from init or server data.
        """
        # from_server_data() may not convert inner objects
        c_obj = [
            (
                x
                if isinstance(x, GuardInterventionCondition)
                else GuardInterventionCondition(**x)  # type: ignore[arg-type]
            )
            for x in conditions
        ]
        a_list = [ModerationGuardAction(x) for x in allowed_actions]
        self.action = ModerationGuardAction(action)
        self.conditions = c_obj
        self.modify_message = modify_message
        self.allowed_actions = a_list
        self.condition_logic = ModerationGuardConditionLogic(condition_logic)  # type: ignore[arg-type]

    def to_dict(self: GuardInterventionForTemplate) -> Dict[str, str]:
        return {
            "action": self.action.value,
            "conditions": [x.to_dict() for x in self.conditions],  # type: ignore[dict-item]
            "modifyMessage": self.modify_message,
            "allowedActions": [x.value for x in self.allowed_actions],  # type: ignore[dict-item]
            "conditionLogic": self.condition_logic.value,
        }

    @classmethod
    def ensure_object(
        cls, maybe_dict: Dict[str, str] | GuardInterventionForTemplate
    ) -> GuardInterventionForTemplate:
        """intervention may arrive as an object, or as a dict. Return an object."""
        if isinstance(maybe_dict, GuardInterventionForTemplate):
            return maybe_dict
        elif isinstance(maybe_dict, dict):
            return GuardInterventionForTemplate(**maybe_dict)
        else:
            raise ValueError("Expected GuardInterventionForTemplate object or dictionary")


class GuardInterventionForConfiguration(APIObject):
    """
    Defines the intervention conditions and actions a guard can take.
    Configuration schema differs slightly from template because changes were requested after
    templates were baked in.
    """

    _converter = t.Dict(
        {
            t.Key("action"): t.Enum(*[e.value for e in ModerationGuardAction]),
            t.Key("conditions"): t.List(GuardInterventionCondition.schema),
            t.Key("message"): String(allow_blank=True),
            t.Key("allowed_actions"): t.List(t.Enum(*[e.value for e in ModerationGuardAction])),
            t.Key("condition_logic", optional=True): t.Enum(
                *[e.value for e in ModerationGuardConditionLogic]
            ),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self: GuardInterventionForConfiguration, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "{}(action={!r})".format(
            self.__class__.__name__,
            self.action,
        )

    def _set_values(
        self: GuardInterventionForConfiguration,
        action: ModerationGuardAction | str,
        conditions: list[GuardInterventionCondition | str],
        message: str,
        allowed_actions: list[ModerationGuardAction | str],
        condition_logic: Optional[
            ModerationGuardConditionLogic
        ] = ModerationGuardConditionLogic.ANY,
    ) -> None:
        """
        Initialize an intervention object based on values from init or server data.
        """
        # from_server_data() may not convert inner objects
        c_obj = [
            (
                x
                if isinstance(x, GuardInterventionCondition)
                else GuardInterventionCondition(**x)  # type: ignore[arg-type]
            )
            for x in conditions
        ]
        a_list = [ModerationGuardAction(x) for x in allowed_actions]
        self.action = ModerationGuardAction(action)
        self.conditions = c_obj
        self.message = message
        self.allowed_actions = a_list
        self.condition_logic = ModerationGuardConditionLogic(condition_logic)  # type: ignore[arg-type]

    def to_dict(self: GuardInterventionForConfiguration) -> Dict[str, str]:
        return {
            "action": self.action.value,
            "conditions": [x.to_dict() for x in self.conditions],  # type: ignore[dict-item]
            "message": self.message,
            "allowedActions": [x.value for x in self.allowed_actions],  # type: ignore[dict-item]
            "conditionLogic": self.condition_logic.value,
        }

    @classmethod
    def ensure_object(
        cls, maybe_dict: Dict[str, str] | GuardInterventionForConfiguration
    ) -> GuardInterventionForConfiguration:
        """intervention may arrive as an object, or as a dict. Return an object."""
        if isinstance(maybe_dict, GuardInterventionForConfiguration):
            return maybe_dict
        elif isinstance(maybe_dict, dict):
            return GuardInterventionForConfiguration(**maybe_dict)
        else:
            raise ValueError("Expected GuardInterventionForConfiguration object or dictionary")



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import cast, Optional, TYPE_CHECKING

import trafaret as t

from datarobot._compat import String
from datarobot.enums import DEFAULT_MAX_WAIT, JOB_TYPE, MODEL_JOB_STATUS

from .. import errors
from ..utils import from_api, retry
from .job import AbstractSpecificJob

if TYPE_CHECKING:
    from datarobot.models.job import Job
    from datarobot.models.model import Model


def wait_for_async_model_creation(
    project_id: str, model_job_id: str, max_wait: int = DEFAULT_MAX_WAIT
) -> Model:
    """
    Given a Project id and ModelJob id poll for status of process
    responsible for model creation until model is created.

    Parameters
    ----------
    project_id : str
        The identifier of the project

    model_job_id : str
        The identifier of the ModelJob

    max_wait : Optional[int]
        Time in seconds after which model creation is considered
        unsuccessful

    Returns
    -------
    model : Model
        Newly created model

    Raises
    ------
    AsyncModelCreationError
        Raised if status of fetched ModelJob object is ``error``
    AsyncTimeoutError
        Model wasn't created in time, specified by ``max_wait`` parameter
    """
    # Note: Don't need this in 3.0. (Use `Job.get_result_when_complete` method instead.)
    for _ in retry.wait(max_wait):
        try:
            model_job = ModelJob.get(project_id, model_job_id)
        except errors.PendingJobFinished:
            return ModelJob.get_model(project_id, model_job_id)
        if model_job.status == MODEL_JOB_STATUS.ERROR:
            e_msg = "Model creation unsuccessful"
            raise errors.AsyncModelCreationError(e_msg)

    timeout_msg = f"Model creation timed out in {max_wait} seconds"
    raise errors.AsyncTimeoutError(timeout_msg)


class ModelJob(AbstractSpecificJob):
    """Tracks asynchronous work being done within a project

    Attributes
    ----------
    id : int
        the id of the job
    project_id : str
        the id of the project the job belongs to
    status : str
        the status of the job - will be one of ``datarobot.enums.QUEUE_STATUS``
    job_type : str
        what kind of work the job is doing - will be 'model' for modeling jobs
    is_blocked : bool
        if true, the job is blocked (cannot be executed) until its dependencies are resolved
    sample_pct : float
        the percentage of the project's dataset used in this modeling job
    model_type : str
        the model this job builds (e.g. 'Nystroem Kernel SVM Regressor')
    processes : List[str]
        the processes used by the model
    featurelist_id : str
        the id of the featurelist used in this modeling job
    blueprint : Blueprint
        the blueprint used in this modeling job
    """

    _extra_fields = frozenset(["sample_pct", "model_type", "processes", "featurelist_id"])

    _converter_extra = t.Dict(
        {
            t.Key("sample_pct", optional=True): t.Float,
            t.Key("model_type", optional=True): String,
            t.Key("processes", optional=True): t.List(String),
            t.Key("featurelist_id", optional=True): String,
        }
    )

    def __repr__(self) -> str:
        return f"ModelJob({self.model_type}, status={self.status})"  # pragma:no cover

    def __init__(self, data, completed_resource_url: Optional[str] = None) -> None:
        super().__init__(data, completed_resource_url=completed_resource_url)
        from . import Blueprint  # pylint: disable=import-outside-toplevel,cyclic-import

        data = from_api(data)

        # ConstructBlueprint
        bp_data = {
            "id": data.get("blueprint_id"),
            "processes": data.get("processes"),
            "model_type": data.get("model_type"),
            "project_id": data.get("project_id"),
        }
        self.blueprint = Blueprint.from_data(bp_data)

    @classmethod
    def _job_type(cls) -> str:
        return cast(str, JOB_TYPE.MODEL)

    @classmethod
    def from_job(cls, job: Job) -> ModelJob:
        """Transforms a generic Job into a ModelJob

        Parameters
        ----------
        job: Job
            A generic job representing a ModelJob

        Returns
        -------
        model_job: ModelJob
            A fully populated ModelJob with all the details of the job

        Raises
        ------
        ValueError:
            If the generic Job was not a model job, e.g. job_type != JOB_TYPE.MODEL
        """
        return super().from_job(job)

    @classmethod
    def get(  # pylint: disable=arguments-renamed
        cls, project_id: str, model_job_id: str
    ) -> ModelJob:
        """
        Fetches one ModelJob. If the job finished, raises PendingJobFinished
        exception.

        Parameters
        ----------
        project_id : str
            The identifier of the project the model belongs to
        model_job_id : str
            The identifier of the model_job

        Returns
        -------
        model_job : ModelJob
            The pending ModelJob

        Raises
        ------
        PendingJobFinished
            If the job being queried already finished, and the server is
            re-routing to the finished model.
        AsyncFailureError
            Querying this resource gave a status code other than 200 or 303
        """
        return super().get(project_id, model_job_id)

    @classmethod
    def _job_path(cls, project_id: str, job_id: str) -> str:
        return f"projects/{project_id}/modelJobs/{job_id}/"

    @classmethod
    def get_model(cls, project_id: str, model_job_id: str) -> Model:
        """
        Fetches a finished model from the job used to create it.

        Parameters
        ----------
        project_id : str
            The identifier of the project the model belongs to
        model_job_id : str
            The identifier of the model_job

        Returns
        -------
        model : Model
            The finished model

        Raises
        ------
        JobNotFinished
            If the job has not finished yet
        AsyncFailureError
            Querying the model_job in question gave a status code other than 200 or
            303
        """
        # Note: Don't need this in 3.0. (Use `Job.get_result` method instead.)
        from . import Model  # pylint: disable=import-outside-toplevel,cyclic-import

        url = cls._job_path(project_id, model_job_id)
        response = cls._client.get(url, allow_redirects=False)
        if response.status_code == 200:
            data = response.json()
            raise errors.JobNotFinished("Pending job status is {}".format(data["status"]))
        elif response.status_code == 303:
            location = response.headers["Location"]
            return Model.from_location(location)
        else:
            e_msg = "Server unexpectedly returned status code {}"
            raise errors.AsyncFailureError(e_msg.format(response.status_code))
